import hashlib
import json

from django.utils.timezone import now
from eveuniverse.models import EveSolarSystem

from app_utils.testing import NoSocketsTestCase

from memberaudit.models import Character, Location
from memberaudit.tests.testdata.factories import (
    create_character_location,
    create_character_update_status,
)
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse
from memberaudit.tests.testdata.load_locations import load_locations
from memberaudit.tests.utils import create_memberaudit_character


class TestCharacterUpdateStatus(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character_1001 = create_memberaudit_character(1001)
        cls.content = {"alpha": 1, "bravo": 2}

    def test_str(self):
        # given
        status = create_character_update_status(
            character=self.character_1001, section=Character.UpdateSection.ASSETS
        )
        # when/then
        self.assertEqual(str(status), f"{self.character_1001}-assets")

    def test_reset_1(self):
        # given
        status = create_character_update_status(
            character=self.character_1001, is_success=True, error_message="abc"
        )
        # when
        status.reset()
        # then
        status.refresh_from_db()
        self.assertIsNone(status.is_success)
        self.assertEqual(status.error_message, "")

    def test_reset_2(self):
        # given
        status = create_character_update_status(
            character=self.character_1001, is_success=True, error_message="abc"
        )
        # when
        status.reset()
        # then
        status.refresh_from_db()
        self.assertIsNone(status.is_success)
        self.assertEqual(status.error_message, "")

    def test_has_changed_1(self):
        """When hash is different, then return True"""
        status = create_character_update_status(
            character=self.character_1001, content_hash_1="abc"
        )
        self.assertTrue(status.has_changed(self.content))

    def test_has_changed_2(self):
        """When no hash exists, then return True"""
        status = create_character_update_status(
            character=self.character_1001, content_hash_1=""
        )
        self.assertTrue(status.has_changed(self.content))

    def test_has_changed_3a(self):
        """When hash is equal, then return False"""
        status = create_character_update_status(
            character=self.character_1001,
            content_hash_1=hashlib.md5(
                json.dumps(self.content).encode("utf-8")
            ).hexdigest(),
        )
        self.assertFalse(status.has_changed(self.content))

    def test_has_changed_3b(self):
        """When hash is equal, then return False"""
        status = create_character_update_status(
            character=self.character_1001,
            content_hash_2=hashlib.md5(
                json.dumps(self.content).encode("utf-8")
            ).hexdigest(),
        )
        self.assertFalse(status.has_changed(content=self.content, hash_num=2))

    def test_has_changed_3c(self):
        """When hash is equal, then return False"""
        status = create_character_update_status(
            character=self.character_1001,
            content_hash_3=hashlib.md5(
                json.dumps(self.content).encode("utf-8")
            ).hexdigest(),
        )
        self.assertFalse(status.has_changed(content=self.content, hash_num=3))

    def test_is_updating_1(self):
        """When run_started_at exist and run_finished_at does not exist, return True"""
        status = create_character_update_status(
            character=self.character_1001, run_started_at=now(), run_finished_at=None
        )
        self.assertTrue(status.is_updating)

    def test_is_updating_2(self):
        """When run_started_at and run_finished_at does not exist, return False"""
        status = create_character_update_status(
            character=self.character_1001, run_started_at=None, run_finished_at=None
        )
        self.assertFalse(status.is_updating)


class TestCharacterLocation(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.character = create_memberaudit_character(1001)

    def test_should_return_location_when_it_exists(self):
        # given
        location: Location = Location.objects.get(id=60003760)
        obj = create_character_location(self.character, location=location)
        # when
        result = obj.location_safe()
        # then
        self.assertEqual(result, location)

    def test_should_return_return_solar_system_when_location_does_not_exist(self):
        # given
        eve_solar_system = EveSolarSystem.objects.get(name="Amamake")
        obj = create_character_location(
            self.character, eve_solar_system=eve_solar_system, location=None
        )
        # when
        result = obj.location_safe()
        # then
        self.assertEqual(result.id, eve_solar_system.id)
        self.assertTrue(result.is_solar_system)



import datetime as dt
from unittest.mock import Mock, patch

from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase, override_settings
from django.urls import reverse
from django.utils.timezone import now
from eveuniverse.models import EveType

from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import create_user_from_evecharacter, generate_invalid_pk

from memberaudit.core import player_count
from memberaudit.models import Character
from memberaudit.tests.testdata.factories import (
    create_character_from_user,
    create_character_mining_ledger_entry,
    create_character_skillpoints,
    create_character_wallet_balance,
    create_character_wallet_journal_entry,
    create_compliance_group,
    create_eve_market_price,
)
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse
from memberaudit.tests.utils import (
    add_auth_character_to_user,
    add_memberaudit_character_to_user,
    create_memberaudit_character,
    create_user_from_evecharacter_with_access,
)
from memberaudit.views.launcher import (
    _dashboard_panel,
    add_character,
    index,
    launcher,
    player_count_data,
    remove_character,
    share_character,
    unshare_character,
)

MODULE_PATH = "memberaudit.views.launcher"


class TestCharacterViews(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_entities()
        cls.user, _ = create_user_from_evecharacter(
            1001, permissions=["memberaudit.basic_access", "memberaudit.reports_access"]
        )

    def test_can_open_index_view(self):
        request = self.factory.get(reverse("memberaudit:index"))
        request.user = self.user
        response = index(request)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("memberaudit:launcher"))

    def test_can_open_launcher_view_1(self):
        """user with main"""
        request = self.factory.get(reverse("memberaudit:launcher"))
        request.user = self.user
        response = launcher(request)
        self.assertEqual(response.status_code, 200)

    def test_can_open_launcher_view_2(self):
        """user without main"""
        user = AuthUtils.create_user("John Doe")
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.basic_access", user
        )

        request = self.factory.get(reverse("memberaudit:launcher"))
        request.user = user
        response = launcher(request)
        self.assertEqual(response.status_code, 200)


@patch(MODULE_PATH + ".messages")
@patch(MODULE_PATH + ".tasks")
@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestAddCharacter(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_entities()
        create_compliance_group()

    def _add_character(self, user, token):
        request = self.factory.get(reverse("memberaudit:add_character"))
        request.user = user
        request.token = token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        orig_view = add_character.__wrapped__.__wrapped__.__wrapped__
        return orig_view(request, token)

    def test_should_add_character(self, mock_tasks, mock_messages):
        # given
        user, _ = create_user_from_evecharacter(
            1001,
            permissions=["memberaudit.basic_access"],
            scopes=Character.get_esi_scopes(),
        )
        token = user.token_set.get(character_id=1001)
        # when
        response = self._add_character(user, token)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("memberaudit:launcher"))
        self.assertTrue(mock_tasks.update_character.apply_async.called)
        self.assertTrue(mock_tasks.update_compliance_groups_for_user.apply_async.called)
        self.assertTrue(mock_messages.success.called)
        self.assertTrue(
            Character.objects.filter(eve_character__character_id=1001).exists()
        )

    def test_should_reenable_disabled_character(self, mock_tasks, mock_messages):
        # given
        character_1001 = create_memberaudit_character(1001)
        character_1001.is_disabled = True
        character_1001.save()
        user = character_1001.character_ownership.user
        token = user.token_set.get(character_id=1001)
        # when
        response = self._add_character(user, token)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("memberaudit:launcher"))
        self.assertTrue(mock_tasks.update_character.apply_async.called)
        self.assertTrue(mock_tasks.update_compliance_groups_for_user.apply_async.called)
        self.assertTrue(mock_messages.success.called)
        character_1001.refresh_from_db()
        self.assertFalse(character_1001.is_disabled)


@patch(MODULE_PATH + ".messages")
@patch(MODULE_PATH + ".tasks")
@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestRemoveCharacter(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_entities()
        create_compliance_group()
        cls.user, _ = create_user_from_evecharacter_with_access(1001)

    def _remove_character(self, user, character_pk):
        request = self.factory.get(
            reverse("memberaudit:remove_character", args=[character_pk])
        )
        request.user = user
        return remove_character(request, character_pk)

    def test_should_remove_character_without_notification(
        self, mock_tasks, mock_messages
    ):
        # given
        character = create_character_from_user(self.user)
        user = character.eve_character.character_ownership.user
        auditor_character = create_memberaudit_character(1003)
        auditor = auditor_character.eve_character.character_ownership.user
        AuthUtils.add_permissions_to_user_by_name(
            (
                "memberaudit.notified_on_character_removal",
                "memberaudit.view_same_corporation",
            ),
            auditor,
        )
        # when
        response = self._remove_character(user, character.pk)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("memberaudit:launcher"))
        self.assertFalse(Character.objects.filter(pk=character.pk).exists())
        self.assertTrue(mock_tasks.update_compliance_groups_for_user.apply_async.called)
        self.assertTrue(mock_messages.success.called)
        self.assertEqual(auditor.notification_set.count(), 0)

    def test_should_remove_character_with_notification(self, mock_tasks, mock_messages):
        # given
        character = create_character_from_user(self.user)
        user = character.eve_character.character_ownership.user
        AuthUtils.add_permission_to_user_by_name("memberaudit.share_characters", user)

        auditor_character = create_memberaudit_character(1002)
        auditor = auditor_character.eve_character.character_ownership.user
        AuthUtils.add_permissions_to_user_by_name(
            (
                "memberaudit.notified_on_character_removal",
                "memberaudit.view_same_corporation",
            ),
            auditor,
        )
        # when
        response = self._remove_character(user, character.pk)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("memberaudit:launcher"))
        self.assertFalse(Character.objects.filter(pk=character.pk).exists())
        self.assertTrue(mock_tasks.update_compliance_groups_for_user.apply_async.called)
        self.assertTrue(mock_messages.success.called)

        expected_removal_notification_title = (
            "Member Audit: Character has been removed!"
        )
        expected_removal_notification_message = (
            "Bruce_Wayne has removed character Bruce Wayne"
        )
        latest_auditor_notification = auditor.notification_set.order_by("-pk")[0]
        self.assertEqual(
            latest_auditor_notification.title, expected_removal_notification_title
        )
        self.assertEqual(
            latest_auditor_notification.message, expected_removal_notification_message
        )
        self.assertEqual(latest_auditor_notification.level, "info")

    def test_should_not_remove_character_from_another_user(
        self, mock_tasks, mock_messages
    ):
        # given
        character_1001 = create_character_from_user(self.user)
        user_1002, _ = create_user_from_evecharacter_with_access(1002)
        # when
        response = self._remove_character(user_1002, character_1001.pk)
        # then
        self.assertEqual(response.status_code, 403)
        self.assertTrue(Character.objects.filter(pk=character_1001.pk).exists())
        self.assertFalse(
            mock_tasks.update_compliance_groups_for_user.apply_async.called
        )
        self.assertFalse(mock_messages.success.called)

    def test_should_respond_with_not_found_for_invalid_characters(
        self, mock_tasks, mock_messages
    ):
        # given
        character = create_character_from_user(self.user)
        user = character.eve_character.character_ownership.user
        invalid_character_pk = generate_invalid_pk(Character)
        # when
        response = self._remove_character(user, invalid_character_pk)
        # then
        self.assertEqual(response.status_code, 404)
        self.assertTrue(Character.objects.filter(pk=character.pk).exists())
        self.assertFalse(
            mock_tasks.update_compliance_groups_for_user.apply_async.called
        )
        self.assertFalse(mock_messages.success.called)


class TestShareCharacter(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_entities()

    def setUp(self) -> None:
        self.character_1001 = create_memberaudit_character(1001)
        self.user_1001 = self.character_1001.eve_character.character_ownership.user
        self.user_1001 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.share_characters", self.user_1001
        )

        self.character_1002 = create_memberaudit_character(1002)
        self.user_1002 = self.character_1002.eve_character.character_ownership.user
        self.user_1002 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.share_characters", self.user_1002
        )

    def test_normal(self):
        request = self.factory.get(
            reverse("memberaudit:share_character", args=[self.character_1001.pk])
        )
        request.user = self.user_1001
        response = share_character(request, self.character_1001.pk)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("memberaudit:launcher"))
        self.assertTrue(Character.objects.get(pk=self.character_1001.pk).is_shared)

    def test_no_permission_1(self):
        """
        when user does not have any permissions
        then redirect to login
        """
        user = AuthUtils.create_user("John Doe")
        request = self.factory.get(
            reverse("memberaudit:share_character", args=[self.character_1001.pk])
        )
        request.user = user
        response = share_character(request, self.character_1001.pk)
        self.assertEqual(response.status_code, 302)
        self.assertIn(reverse("login"), response.url)

    def test_no_permission_2(self):
        """
        when user does has basic_access only
        then redirect to login
        """
        user = AuthUtils.create_user("John Doe")
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.basic_access", user
        )
        request = self.factory.get(
            reverse("memberaudit:share_character", args=[self.character_1001.pk])
        )
        request.user = user
        response = share_character(request, self.character_1001.pk)
        self.assertEqual(response.status_code, 302)
        self.assertIn(reverse("login"), response.url)

    def test_no_permission_3(self):
        request = self.factory.get(
            reverse("memberaudit:share_character", args=[self.character_1001.pk])
        )
        request.user = self.user_1002
        response = share_character(request, self.character_1001.pk)
        self.assertEqual(response.status_code, 403)
        self.assertFalse(Character.objects.get(pk=self.character_1001.pk).is_shared)

    def test_not_found(self):
        invalid_character_pk = generate_invalid_pk(Character)
        request = self.factory.get(
            reverse("memberaudit:share_character", args=[invalid_character_pk])
        )
        request.user = self.user_1001
        response = share_character(request, invalid_character_pk)
        self.assertEqual(response.status_code, 404)
        self.assertFalse(Character.objects.get(pk=self.character_1001.pk).is_shared)


class TestUnshareCharacter(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_entities()

    def setUp(self) -> None:
        self.character_1001 = create_memberaudit_character(1001)
        self.character_1001.is_shared = True
        self.character_1001.save()
        self.user_1001 = self.character_1001.eve_character.character_ownership.user

        self.character_1002 = create_memberaudit_character(1002)
        self.user_1002 = self.character_1002.eve_character.character_ownership.user

    def test_normal(self):
        request = self.factory.get(
            reverse("memberaudit:unshare_character", args=[self.character_1001.pk])
        )
        request.user = self.user_1001
        response = unshare_character(request, self.character_1001.pk)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("memberaudit:launcher"))
        self.assertFalse(Character.objects.get(pk=self.character_1001.pk).is_shared)

    def test_no_permission(self):
        request = self.factory.get(
            reverse("memberaudit:unshare_character", args=[self.character_1001.pk])
        )
        request.user = self.user_1002
        response = unshare_character(request, self.character_1001.pk)
        self.assertEqual(response.status_code, 403)
        self.assertTrue(Character.objects.get(pk=self.character_1001.pk).is_shared)

    def test_not_found(self):
        invalid_character_pk = generate_invalid_pk(Character)
        request = self.factory.get(
            reverse("memberaudit:unshare_character", args=[invalid_character_pk])
        )
        request.user = self.user_1001
        response = unshare_character(request, invalid_character_pk)
        self.assertEqual(response.status_code, 404)
        self.assertTrue(Character.objects.get(pk=self.character_1001.pk).is_shared)


class TestDashboardPanel(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.factory = RequestFactory()

    def setUp(self) -> None:
        player_count.clear_cache()

    def test_user_with_complete_data(self):
        # given
        character_1001 = create_memberaudit_character(1001)
        user = character_1001.user
        character_1002 = add_memberaudit_character_to_user(user, 1002)
        add_auth_character_to_user(user, 1003)

        create_character_skillpoints(character_1001, total=1_000)
        create_character_skillpoints(character_1002, total=3_000)

        create_character_wallet_balance(character_1001, total=10_000)
        create_character_wallet_balance(character_1002, total=5_000)

        today = now().date()
        ore_type = EveType.objects.get(name="Veldspar")
        create_eve_market_price(eve_type=ore_type, average_price=100)
        create_character_mining_ledger_entry(
            character_1001,
            eve_type=ore_type,
            quantity=4,
            date=today - dt.timedelta(days=1),
        )
        create_character_mining_ledger_entry(
            character_1001, eve_type=ore_type, quantity=3, date=today
        )
        create_character_mining_ledger_entry(
            character_1002, eve_type=ore_type, quantity=2, date=today
        )
        not_this_month = now() - dt.timedelta(days=40)
        create_character_mining_ledger_entry(
            character_1002, eve_type=ore_type, quantity=2, date=not_this_month
        )
        create_character_wallet_journal_entry(
            character_1001, amount=4_000, ref_type="bounty_prizes", date=now()
        )
        create_character_wallet_journal_entry(
            character_1001, amount=3_000, ref_type="bounty_prizes", date=now()
        )
        create_character_wallet_journal_entry(
            character_1002, amount=2_000, ref_type="bounty_prizes", date=now()
        )
        not_this_month = now() - dt.timedelta(days=40)
        create_character_wallet_journal_entry(
            character_1002,
            amount=2_000,
            ref_type="bounty_prizes",
            date=not_this_month,
        )

        request = self.factory.get("/")
        request.user = user

        # when
        context = _dashboard_panel(request)

        # then
        self.assertEqual(context["registered_count"], 2)
        self.assertEqual(context["known_characters_count"], 3)
        self.assertEqual(context["registered_percent"], 67)
        self.assertEqual(context["total_wallet_isk"], 15_000)
        self.assertEqual(context["total_ratted_isk"], 9_000)
        self.assertEqual(context["total_mined_isk"], 900.0)
        self.assertEqual(context["total_character_skillpoints"], 4_000)

    def test_user_with_memberaudit_character_and_no_data(self):
        # given
        character_1001 = create_memberaudit_character(1001)
        user = character_1001.user
        request = self.factory.get("/")
        request.user = user

        # when
        context = _dashboard_panel(request)

        # then
        self.assertEqual(context["registered_count"], 1)
        self.assertEqual(context["known_characters_count"], 1)
        self.assertEqual(context["registered_percent"], 100)
        self.assertIsNone(context["total_wallet_isk"])
        self.assertEqual(context["total_ratted_isk"], 0)
        self.assertEqual(context["total_mined_isk"], 0)
        self.assertIsNone(context["total_character_skillpoints"])

    def test_user_with_memberaudit_character_and_no_current_mining_and_ratting_data(
        self,
    ):
        # given
        character_1001 = create_memberaudit_character(1001)
        user = character_1001.user
        not_this_month = now().date() - dt.timedelta(days=40)
        create_character_mining_ledger_entry(character_1001, date=not_this_month)
        create_character_wallet_journal_entry(
            character_1001, ref_type="bounty_prizes", date=not_this_month
        )

        request = self.factory.get("/")
        request.user = user

        # when
        context = _dashboard_panel(request)

        # then
        self.assertEqual(context["total_ratted_isk"], 0)
        self.assertEqual(context["total_mined_isk"], 0)


@patch(MODULE_PATH + ".player_count.get", spec=True)
class TestPlayerCountData(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_entities()
        character = create_memberaudit_character(1001)
        cls.user = character.user

    def test_should_return_player_count(self, mock_player_count):
        # given
        mock_player_count.return_value = 42
        request = self.factory.get("/")
        request.user = self.user

        # when
        response = player_count_data(request)

        # then
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content.decode("utf-8"), {"player_count": 42})



from app_utils.testing import NoSocketsTestCase

from memberaudit.core.xml_converter import DEFAULT_FONT_SIZE, eve_xml_to_html
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse
from memberaudit.tests.testdata.load_locations import load_locations

MODULE_PATH = "memberaudit.core.xml_converter"


class TestXMLConversion(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()

    def test_should_convert_font_tag(self):
        input = '<font size="13" color="#b3ffffff">Character</font>'
        expected = '<span style="font-size: 13px">Character</span>'
        self.assertHTMLEqual(eve_xml_to_html(input), expected)

    def test_should_remove_loc_tag(self):
        input = "<loc>Character</loc>"
        expected = "Character"
        self.assertHTMLEqual(eve_xml_to_html(input), expected)

    def test_add_target_to_normal_links(self):
        input = (
            '<a href="http://www.google.com" target="_blank">https://www.google.com</a>'
        )
        self.assertHTMLEqual(eve_xml_to_html(input), input)

    def test_should_convert_character_link(self):
        input = '<a href="showinfo:1376//1001">Bruce Wayne</a>'
        expected = '<a href="https://evewho.com/character/1001" target="_blank">Bruce Wayne</a>'
        self.assertHTMLEqual(eve_xml_to_html(input), expected)

    def test_should_convert_corporation_link(self):
        input = '<a href="showinfo:2//2001">Wayne Technologies</a>'
        expected = (
            '<a href="https://evemaps.dotlan.net/corp/Wayne_Technologies" '
            'target="_blank">Wayne Technologies</a>'
        )
        self.assertHTMLEqual(eve_xml_to_html(input), expected)

    def test_should_convert_alliance_link(self):
        input = '<a href="showinfo:16159//3001">Wayne Enterprises</a>'
        expected = (
            '<a href="https://evemaps.dotlan.net/alliance/Wayne_Enterprises" '
            'target="_blank">Wayne Enterprises</a>'
        )
        self.assertHTMLEqual(eve_xml_to_html(input), expected)

    def test_should_convert_solar_system_link(self):
        input = '<a href="showinfo:5//30004984">Abune</a>'
        expected = '<a href="https://evemaps.dotlan.net/system/Abune" target="_blank">Abune</a>'
        self.assertHTMLEqual(eve_xml_to_html(input), expected)

    def test_should_convert_station_link(self):
        input = (
            '<a href="showinfo:52678//60003760">'
            "Jita IV - Moon 4 - Caldari Navy Assembly Plant</a>"
        )
        expected = (
            '<a href="https://evemaps.dotlan.net/station/Jita_IV_-_Moon_4_-_Caldari_Navy_Assembly_Plant" '
            'target="_blank">Jita IV - Moon 4 - Caldari Navy Assembly Plant</a>'
        )
        self.assertHTMLEqual(eve_xml_to_html(input), expected)

    def test_should_convert_kill_link(self):
        input = (
            '<a href="killReport:84900666:9e6fe9e5392ff0cfc6ab956677dbe1deb69c4b04">'
            "Kill: Yuna Kobayashi (Badger)</a>"
        )
        expected = (
            '<a href="https://zkillboard.com/kill/84900666/" '
            'target="_blank">Kill: Yuna Kobayashi (Badger)</a>'
        )
        self.assertHTMLEqual(eve_xml_to_html(input), expected)

    # def test_should_disable_unknown_types(self):
    #     input = '<a href="showinfo:601//30004984">Abune</a>'
    #     expected = '<a href="#">Abune</a>'
    #     self.assertHTMLEqual(eve_xml_to_html(input), expected)

    def test_should_disable_unknown_links(self):
        input = '<a href="unknown">Abune</a>'
        expected = '<a href="#">Abune</a>'
        self.assertHTMLEqual(eve_xml_to_html(input), expected)

    def test_should_set_default_font(self):
        input = 'First<br><font size="20">Second</font>Third'
        expected = (
            '<span style="font-size: 13px">First</span>'
            '<br><span style="font-size: 20px">Second</span>'
            '<span style="font-size: 13px">Third</span>'
        )
        self.assertHTMLEqual(eve_xml_to_html(input, add_default_style=True), expected)

    def test_should_remove_comment(self):
        input = "<u>First<!--<script>badcall();</script>--></u>"
        expected = "<u>First</u>"
        self.assertHTMLEqual(eve_xml_to_html(input), expected)

    def test_should_remove_normal_script(self):
        input = "<b>This is a <script>bad_attempt()</script> at injection</b>"
        expected = "<b>This is a bad_attempt() at injection</b>"
        self.assertHTMLEqual(eve_xml_to_html(input), expected)

    # \ufe64 and \ufe65 are two examples of characters that normalize under NFKC to < >
    def test_should_remove_confusable_script(self):
        input = "<i>Yet another \ufe64script\ufe65attempted()\ufe64/script\ufe65 injection</i>"
        expected = "<i>Yet another attempted() injection</i>"
        self.assertHTMLEqual(eve_xml_to_html(input), expected)

    def test_should_move_valid_children_out(self):
        input = "<u>Test <div><i>it</i> <h1>out</h1></div> okay?</u>"
        expected = "<u>Test <i>it</i> out okay?</u>"
        self.assertHTMLEqual(eve_xml_to_html(input), expected)

    def test_should_use_default_font_size(self):
        input = '<font size="invalid">Test</font>'
        expected = f'<span style="font-size: {DEFAULT_FONT_SIZE}px">Test</span>'
        self.assertHTMLEqual(eve_xml_to_html(input), expected)

    def test_should_use_absolute_font_size(self):
        input = '<font size="-13">Test</font>'
        expected = '<span style="font-size: 13px">Test</span>'
        self.assertHTMLEqual(eve_xml_to_html(input), expected)

    def test_should_handle_valid_or_invalid_protocols(self):
        input = '<a href="unsupported://123456">I should just be text</a><br><a href="https://example.org/">I should be a link</a>'
        expected = 'I should just be text<br><a href="https://example.org/" target="_blank">I should be a link</a>'
        self.assertHTMLEqual(eve_xml_to_html(input), expected)

    def test_should_remove_empty_elements(self):
        input = '<font size="13"><font size="10"></font><a href="https://iforgottext.com/"></a>some text</font>'
        expected = '<span style="font-size: 13px">some text</span>'
        self.assertHTMLEqual(eve_xml_to_html(input), expected)



"""Return current statistics about Member Audit."""

import datetime as dt
import logging

from tqdm import tqdm

from django.contrib.auth.models import User
from django.core.management.base import BaseCommand
from django.db.models import Avg, Count, F, Max, Min

from app_utils.logging import LoggerAddTag

from memberaudit import __title__, app_settings
from memberaudit.constants import IS_TESTING
from memberaudit.helpers import character_section_models
from memberaudit.management.commands._helpers import Table
from memberaudit.models import Character, CharacterUpdateStatus, characters

from . import get_input

logger = LoggerAddTag(logging.getLogger(__name__), __title__)


class Command(BaseCommand):
    help = str(__doc__)

    def handle(self, *args, **options):
        self.stdout.write("1 - Settings")
        self.stdout.write("2 - Object counts")
        self.stdout.write("3 - Stale minutes")
        self.stdout.write("4 - Update statistics")

        result = get_input("Enter 1 - 4 to choose a menu or any other input to exit? ")
        self.stdout.write("")

        if result not in ["1", "2", "3", "4"]:
            self.stdout.write(self.style.WARNING("Aborted"))

        if result == "1":
            data = _fetch_settings()
            self._output_section(data, "settings")

        elif result == "2":
            data = _calc_object_counts()
            self._output_section(data, "object counts", "right")

        elif result == "3":
            data = _fetch_stale_minutes()
            self._output_section(data, "stale minutes", "right")

        elif result == "4":
            self.stdout.write("Calculating...\r", ending="")
            data = _calc_update_stats()

            self._write_title("update statistic")
            table = Table(default_alignment=Table.Alignment.RIGHT)
            table.set_data(data)
            table.set_alignment(0, Table.Alignment.LEFT)
            table.write(self.stdout)

    def _write_title(self, text: str):
        output = text.title()
        self.stdout.write(self.style.SUCCESS(f"{output:<20}"))
        self.stdout.write("")

    def _output_section(self, data: dict, title: str, alignment: str = "left"):
        self._write_title(title)

        if alignment == "left":
            alignment_symbol = "<"
        elif alignment == "right":
            alignment_symbol = ">"
        else:
            raise NotImplementedError(f"Unknown alignment: {alignment}")

        data_sorted = dict(sorted(data.items()))
        data_formatted = {
            label: self._format_value(value) for label, value in data_sorted.items()
        }
        label_width = max(len(label) for label in data_formatted) + 1
        value_width = max(len(value) for value in data_formatted.values()) + 1

        for label, value in data_formatted.items():
            self.stdout.write(
                f"  {label:<{label_width}}: {value:{alignment_symbol}{value_width}}"
            )

    def _format_value(self, value):
        return f"{value:,}" if isinstance(value, (int, float)) else str(value)


def _fetch_stale_minutes():
    return dict(characters.section_time_until_stale)


def _calc_object_counts():
    work = []
    for model_class in character_section_models():
        name = str(model_class._meta.verbose_name_plural)
        query = model_class.objects.all()
        work.append((name, query))

    characters_enabled = Character.objects.filter(is_disabled=False)
    work.append(("characters enabled", characters_enabled))

    user_query = User.objects.filter(
        character_ownerships__character__memberaudit_character__isnull=False
    ).distinct()
    work.append(("users with access", user_query))

    object_counts = {
        name: query.count()
        for name, query, in tqdm(
            work, desc="Calculating object counts", leave=False, disable=IS_TESTING
        )
    }
    return object_counts


def _calc_update_stats():
    sections = {section: {"section": section} for section in Character.UpdateSection}

    durations = (
        CharacterUpdateStatus.objects.filter(
            is_success=True,
            update_started_at__isnull=False,
            update_finished_at__isnull=False,
        )
        .annotate(duration=F("update_finished_at") - F("update_started_at"))
        .values("section")
        .annotate(duration_min=Min("duration"))
        .annotate(duration_avg=Avg("duration"))
        .annotate(duration_max=Max("duration"))
        .annotate(sample_size=Count("pk"))
        .values(
            "section", "duration_min", "duration_avg", "duration_max", "sample_size"
        )
    )
    durations_mapped = {
        Character.UpdateSection(obj["section"]): obj for obj in durations
    }

    duration_fields = ("duration_min", "duration_avg", "duration_max", "sample_size")
    for section in sections:
        try:
            obj = durations_mapped[section]
        except KeyError:
            section_durations = {field: None for field in duration_fields}
        else:
            section_durations = {
                field: _convert_timedelta(obj[field]) for field in duration_fields
            }

        sections[section].update(section_durations)

    return list(sections.values())


def _convert_timedelta(value):
    if isinstance(value, dt.timedelta):
        return value.total_seconds()
    return value


def _fetch_settings():
    settings = {
        name: value
        for name, value in vars(app_settings).items()
        if name.startswith("MEMBERAUDIT_")
        and name
        not in {
            "MEMBERAUDIT_BASE_URL",
            "MEMBERAUDIT_SECTION_STALE_MINUTES_SECTION_DEFAULTS",
        }
    }
    return settings



"""Define admin site for Member Audit."""

# pylint: disable=missing-class-docstring,missing-function-docstring

import datetime as dt
import functools
from typing import Any, List, Optional

from humanize.time import naturaldelta, naturaltime

from django import forms
from django.contrib import admin
from django.contrib.auth.models import Group
from django.core.exceptions import ValidationError
from django.db.models import Case, Max, Prefetch, Q, QuerySet, TextChoices, Value, When
from django.forms.models import BaseInlineFormSet
from django.http.request import HttpRequest
from django.shortcuts import redirect, render
from django.utils.html import format_html
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from eveuniverse.models import EveType

from allianceauth.authentication.models import State

from memberaudit import tasks
from memberaudit.app_settings import MEMBERAUDIT_TASKS_NORMAL_PRIORITY
from memberaudit.constants import EveCategoryId, EveTypeId
from memberaudit.models import (
    Character,
    CharacterUpdateStatus,
    ComplianceGroupDesignation,
    EveShipType,
    EveSkillType,
    Location,
    SkillSet,
    SkillSetGroup,
    SkillSetSkill,
)


class AddDeleteObjects:
    """Mixin for adding the action to delete selected objects.

    Note: One also need to add the new action "delete_objects" to the actions list.
    """

    def get_actions(self, request):
        """Remove the default delete action from the drop-down."""
        actions = super().get_actions(request)
        if "delete_selected" in actions:
            del actions["delete_selected"]
        return actions

    @admin.action(description=_("Delete selected objects"))
    def delete_objects(self, request, queryset):
        if "apply" in request.POST:
            pks = list(queryset.values_list("pk", flat=True))
            model_name = queryset.model.__name__
            tasks.delete_objects.apply_async(
                args=[model_name, pks],
                priority=MEMBERAUDIT_TASKS_NORMAL_PRIORITY,
            )  # type: ignore
            self.message_user(
                request,
                _("Started deleting %d %s objects. This can take a minute.")
                % (len(pks), model_name),
            )
            return redirect(request.get_full_path())

        return render(
            request,
            "admin/memberaudit/confirm_deleting_objects.html",
            {
                "title": _("Are you sure you want to delete these objects?"),
                "queryset": queryset.all(),
                "action": "delete_objects",
            },
        )


class ComplianceGroupDesignationForm(forms.ModelForm):
    class Meta:
        model = ComplianceGroupDesignation
        fields = ("group",)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            self.fields["group"].queryset = Group.objects.filter(
                authgroup__internal=True, compliancegroupdesignation__isnull=True
            ).order_by("name")
        except KeyError:
            pass


@admin.register(ComplianceGroupDesignation)
class ComplianceGroupDesignationAdmin(admin.ModelAdmin):
    form = ComplianceGroupDesignationForm
    ordering = ("group__name",)
    list_display = ("_group_name", "_states")
    list_display_links = None

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related("group").prefetch_related("group__authgroup__states")

    def save_model(self, request, obj, *args, **kwargs) -> None:
        super().save_model(request, obj, *args, **kwargs)
        tasks.add_compliant_users_to_group.apply_async(
            args=[obj.group.pk], priority=MEMBERAUDIT_TASKS_NORMAL_PRIORITY
        )  # type: ignore

    def delete_queryset(self, request, queryset) -> None:
        for obj in queryset:
            tasks.clear_users_from_group.apply_async(
                args=[obj.group.pk], priority=MEMBERAUDIT_TASKS_NORMAL_PRIORITY
            )  # type: ignore
            obj.delete()

    @admin.display(ordering="group__name")
    def _group_name(self, obj) -> str:
        return obj.group.name

    @admin.display(description=_("Restricted to states"))
    def _states(self, obj):
        states = [state.name for state in obj.group.authgroup.states.all()]
        return sorted(states) if states else "-"

    def has_change_permission(self, request, obj=None):
        return False


class EveUniverseEntityModelAdmin(admin.ModelAdmin):
    def has_module_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False

    ordering = ["name"]
    search_fields = ["name"]


@admin.register(EveShipType)
class EveShipTypeAdmin(EveUniverseEntityModelAdmin):
    pass


@admin.register(EveSkillType)
class EveSkillTypeAdmin(EveUniverseEntityModelAdmin):
    pass


class CharacterUpdateStatusAdminInline(admin.TabularInline):
    model = CharacterUpdateStatus
    fields = (
        "section",
        "_is_enabled",
        "_is_success",
        "_is_token_ok",
        "error_message",
        "run_finished_at",
        "_run_duration",
        "update_finished_at",
        "_update_duration",
    )
    readonly_fields = (
        "_is_enabled",
        "_is_success",
        "_is_token_ok",
        "_run_duration",
        "_update_duration",
    )
    ordering = ["section"]

    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    @admin.display(boolean=True)
    def _is_enabled(self, obj: CharacterUpdateStatus) -> bool:
        return obj.is_enabled

    @admin.display(boolean=True)
    def _is_success(self, obj: CharacterUpdateStatus) -> bool:
        if not obj.is_enabled:
            return None
        return obj.is_success

    @admin.display(boolean=True)
    def _is_token_ok(self, obj: CharacterUpdateStatus) -> bool:
        return not obj.has_token_error

    @admin.display
    def _run_duration(self, obj: CharacterUpdateStatus) -> float:
        return self._calc_duration(obj.run_started_at, obj.run_finished_at)

    @admin.display
    def _update_duration(self, obj: CharacterUpdateStatus) -> float:
        return self._calc_duration(obj.update_started_at, obj.update_finished_at)

    @staticmethod
    def _calc_duration(
        started_at: dt.datetime, finished_at: dt.datetime
    ) -> dt.timedelta:
        if not started_at or not finished_at:
            return "-"

        return naturaldelta(finished_at - started_at)


class CharacterUpdateStatusListFilter(admin.SimpleListFilter):
    """Custom filter for update status with counts."""

    title = _("update status")
    parameter_name = "total_update_status"

    def lookups(self, request, model_admin):
        qs = model_admin.get_queryset(request)
        counts = []
        for status in Character.TotalUpdateStatus:
            counts.append((status, qs.filter(total_update_status=status.value).count()))
        result = tuple(
            (
                (status.value, status.label.title() + f" ({count:,})")
                for status, count in counts
            )
        )
        return result

    def queryset(self, request, queryset):
        for value in Character.TotalUpdateStatus.values:
            if self.value() == value:
                return queryset.filter(total_update_status=value)
        return queryset


class CharacterStateListFilter(admin.SimpleListFilter):
    """Custom state filter to include filtering of characters without main."""

    title = _("state")
    parameter_name = "state"
    _NO_MAIN_KEY = "_NO_MAIN"

    def __init__(self, *args, **kwargs) -> None:
        self._states = State.objects.order_by("-priority").values_list(
            "name", flat=True
        )
        super().__init__(*args, **kwargs)

    def lookups(self, request, model_admin):
        qs = model_admin.get_queryset(request)
        counts = []
        for name in self._states:
            counts.append(
                (
                    name,
                    qs.filter(
                        eve_character__character_ownership__user__profile__state__name=name
                    ).count(),
                )
            )
        result = [(name, name + f" ({count:,})") for name, count in counts]
        count_no_main = qs.filter(
            eve_character__character_ownership__isnull=True
        ).count()
        result.append((self._NO_MAIN_KEY, _("No main") + f" ({count_no_main:,})"))
        return result

    def queryset(self, request, queryset):
        value = self.value()
        if value == self._NO_MAIN_KEY:
            return queryset.filter(eve_character__character_ownership__isnull=True)
        for name in self._states:
            if value == name:
                return queryset.filter(
                    eve_character__character_ownership__user__profile__state__name=name
                )
        return queryset


def generic_action_update_section(
    modeladmin, request, queryset, section: Character.UpdateSection
):
    """Update a section for the selected characters as generic action."""
    for obj in queryset:
        task = getattr(tasks, f"update_character_{section.value}")
        task.apply_async(
            kwargs={"character_pk": obj.pk, "force_update": True},
            priority=MEMBERAUDIT_TASKS_NORMAL_PRIORITY,
        )  # type: ignore

    modeladmin.message_user(
        request,
        _("Started updating %(section)s for %(count)s characters.")
        % {"section": section.label, "count": queryset.count()},
    )


@admin.register(Character)
class CharacterAdmin(AddDeleteObjects, admin.ModelAdmin):
    class Media:
        css = {
            "all": (
                "allianceauth/authentication/css/admin.css",
                "memberaudit/css/admin.css",
            ),
        }

    list_display = (
        "_character_pic",
        "_character",
        "_main",
        "_state",
        "_organization",
        "_created_at",
        "_enabled",
        "_last_update_at",
        "_update_status",
        "_missing_sections",
    )
    list_display_links = (
        "_character_pic",
        "_character",
    )
    list_filter = (
        CharacterUpdateStatusListFilter,
        "is_disabled",
        CharacterStateListFilter,
        "created_at",
        "eve_character__character_ownership__user__profile__main_character__alliance_name",
    )
    list_select_related = (
        "eve_character__character_ownership__user",
        "eve_character__character_ownership__user__profile__main_character",
        "eve_character__character_ownership__user__profile__state",
        "eve_character",
    )
    ordering = ["eve_character__character_name"]
    search_fields = [
        "eve_character__character_name",
        "eve_character__character_ownership__user__profile__main_character__character_name",
        "eve_character__character_ownership__user__profile__main_character__corporation_name",
        "eve_character__character_ownership__user__profile__main_character__alliance_name",
    ]
    exclude = ("mailing_lists",)

    actions = [
        "delete_objects",
        "update_characters",
        "update_assets",
        "update_roles",
        "update_location",
        "update_online_status",
        "enable_characters",
        "disable_characters",
    ]
    inlines = (CharacterUpdateStatusAdminInline,)

    def get_queryset(self, *args, **kwargs):
        qs = super().get_queryset(*args, **kwargs)
        return (
            qs.prefetch_related("update_status_set")
            .annotate(
                last_update_at=Max(
                    "update_status_set__run_finished_at",
                    filter=~Q(update_status_set__section="skill_sets"),
                )
            )
            .annotate_total_update_status()
        )

    def get_actions(self, request):
        """Generate and add generated actions for all sections
        which can be updated through single task.
        """
        actions: dict = super().get_actions(request)
        sections = sorted(list(Character.UpdateSection.enabled_sections()))
        for section in sections:
            action_name = f"update_section_{section}"
            func = functools.partial(generic_action_update_section, section=section)
            description = f"Update {section.label} for selected characters"
            actions[action_name] = (func, action_name, description)

        return dict(sorted(actions.items(), key=lambda item: item[1][2]))

    @admin.display(description="")
    def _character_pic(self, obj: Character):
        character = obj.eve_character
        return format_html(
            '<img src="{}" class="img-circle">', character.portrait_url(size=32)
        )

    @admin.display(ordering="eve_character__character_name", description=_("character"))
    def _character(self, obj: Character) -> str:
        return str(obj.eve_character)

    @admin.display(ordering="is_disabled", boolean=True, description=_("enabled"))
    def _enabled(self, obj: Character) -> bool:
        return not obj.is_disabled

    @admin.display(
        ordering="eve_character__character_ownership__user__profile__main_character",
        description=_("main"),
    )
    def _main(self, obj: Character) -> Optional[str]:
        try:
            name = obj.main_character.character_name
        except AttributeError:
            return None
        return str(name)

    @admin.display(
        ordering="eve_character__character_ownership__user__profile__state__name",
        description=_("state"),
    )
    def _state(self, obj: Character) -> Optional[str]:
        try:
            return str(obj.user.profile.state)
        except AttributeError:
            return None

    @admin.display(
        ordering="eve_character__character_ownership__user__profile__main_character__corporation_name",
        description=_("organization"),
    )
    def _organization(self, obj: Character) -> Optional[str]:
        if not obj.main_character:
            return None
        result = obj.main_character.corporation_name
        if result and obj.main_character.alliance_ticker:
            result += f" [{obj.main_character.alliance_ticker}]"
        return result

    @admin.display(ordering="total_update_status", description=_("update status"))
    def _update_status(self, obj: Character):
        update_status_obj = Character.TotalUpdateStatus(obj.total_update_status)
        label = update_status_obj.label.title()
        css_class = update_status_obj.bootstrap_style_class()
        description = update_status_obj.description()
        return format_html(
            '<span class="{}" title="{}">{}</span>', css_class, description, label
        )

    @admin.display(ordering="created_at", description=_("created"))
    def _created_at(self, obj: Character):
        return obj.created_at

    @admin.display(ordering="last_update_at", description=_("last update run"))
    def _last_update_at(self, obj: Character):
        return naturaltime(obj.last_update_at) if obj.last_update_at else "-"

    def _missing_sections(self, obj):
        existing = {status.section for status in obj.update_status_set.all()}
        enabled_sections = Character.UpdateSection.enabled_sections()
        missing_sections = enabled_sections.difference(existing)
        if missing_sections:
            return sorted(obj.label for obj in missing_sections)
        return None

    @admin.action(description=_("Update all for selected characters"))
    def update_characters(self, request, queryset):
        for obj in queryset:
            tasks.update_character.apply_async(
                kwargs={
                    "character_pk": obj.pk,
                    "force_update": True,
                    "ignore_stale": True,
                },
                priority=MEMBERAUDIT_TASKS_NORMAL_PRIORITY,
            )  # type: ignore

        self.message_user(
            request, _("Started updating %d characters.") % queryset.count()
        )

    @admin.action(
        description=_("Enable selected characters and reset token notifications")
    )
    def enable_characters(self, request, queryset):
        pks = list(queryset.values_list("pk", flat=True))
        queryset.filter(pk__in=pks).update(
            is_disabled=False, token_error_notified_at=None
        )
        self.message_user(request, _("Enabled %d characters.") % len(pks))

    @admin.action(description=_("Disable selected characters"))
    def disable_characters(self, request, queryset):
        pks = list(queryset.values_list("pk", flat=True))
        queryset.filter(pk__in=pks).update(is_disabled=True)
        self.message_user(request, _("Disabled %d characters.") % len(pks))

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False


class LocationCategory(TextChoices):
    SOLAR_SYSTEM = "solar_system", "Solar System"
    STATION = "station", "Station"
    STRUCTURE = "structure", "Structure"
    UNKNOWN = "unknown", "Unknown"


class LocationCategoryListFilter(admin.SimpleListFilter):
    title = _("category")
    parameter_name = "category"

    def lookups(self, request, model_admin):
        return LocationCategory.choices

    def queryset(self, request, queryset):
        if value := self.value():
            return queryset.filter(category=value)
        return None


@admin.register(Location)
class LocationAdmin(admin.ModelAdmin):
    list_display = ("id", "_name", "_type", "_group", "_solar_system", "updated_at")
    list_filter = (
        LocationCategoryListFilter,
        ("eve_type__eve_group", admin.RelatedOnlyFieldListFilter),
        "updated_at",
    )
    search_fields = [
        "id",
        "name",
        "eve_solar_system__eve_constellation__eve_region__name",
        "eve_type__name",
    ]
    ordering = ["id"]

    def get_queryset(self, request: HttpRequest) -> QuerySet[Any]:
        qs = super().get_queryset(request)
        qs = qs.select_related(
            "eve_type__eve_group",
            "eve_type",
            "eve_solar_system__eve_constellation__eve_region",
            "eve_solar_system",
        ).annotate(
            category=Case(
                When(
                    eve_type_id=EveTypeId.SOLAR_SYSTEM,
                    then=Value(LocationCategory.SOLAR_SYSTEM),
                ),
                When(
                    eve_type__eve_group__eve_category_id=EveCategoryId.STATION,
                    then=Value(LocationCategory.STATION),
                ),
                When(
                    eve_type__eve_group__eve_category_id=EveCategoryId.STRUCTURE,
                    then=Value(LocationCategory.STRUCTURE),
                ),
                default=Value(LocationCategory.UNKNOWN),
            )
        )
        return qs

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    @admin.display(ordering="name", description=_("name"))
    def _name(self, obj):
        return obj.name_plus

    @admin.display(ordering="eve_solar_system__name", description=_("solar system"))
    def _solar_system(self, obj):
        return obj.eve_solar_system.name if obj.eve_solar_system else None

    @admin.display(ordering="eve_type__name", description=_("type"))
    def _type(self, obj):
        return obj.eve_type.name if obj.eve_type else None

    @admin.display(ordering="eve_type__eve_group__name", description=_("group"))
    def _group(self, obj):
        return obj.eve_type.eve_group.name if obj.eve_type else None


@admin.register(SkillSetGroup)
class SkillSetGroupAdmin(admin.ModelAdmin):
    list_display = ("name", "_skill_sets", "is_doctrine", "is_active")
    list_filter = (
        "is_doctrine",
        "is_active",
        ("skill_sets", admin.RelatedOnlyFieldListFilter),
    )
    ordering = ["name"]
    filter_horizontal = ("skill_sets",)
    readonly_fields = ("last_modified_at", "last_modified_by")
    fields = [
        "name",
        "description",
        "skill_sets",
        "is_doctrine",
        "is_active",
        ("last_modified_at", "last_modified_by"),
    ]

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.prefetch_related(
            Prefetch(
                "skill_sets",
                queryset=SkillSet.objects.order_by("name"),
                to_attr="skill_sets_ordered",
            )
        )

    def _skill_sets(self, obj):
        return format_html(
            "<br>".join([skill_set.name for skill_set in obj.skill_sets_ordered])
        )

    def save_model(self, request, obj, form, change):
        obj.last_modified_by = request.user
        obj.last_modified_at = now()
        super().save_model(request, obj, form, change)


class SkillSetSkillAdminFormSet(BaseInlineFormSet):
    def clean(self):
        super().clean()
        for _form in self.forms:
            try:
                data = self.cleaned_data
            except AttributeError:
                pass
            else:
                for row in data:
                    if (
                        row
                        and row.get("DELETE") is False
                        and not row.get("required_level")
                        and not row.get("recommended_level")
                    ):
                        eve_type = row.get("eve_type")
                        raise ValidationError(
                            _("Skill '%s' must have a level.") % eve_type.name
                        )


class SkillSetSkillAdminInline(admin.TabularInline):
    model = SkillSetSkill
    verbose_name = "skill"
    verbose_name_plural = "skills"
    min_num = 1
    formset = SkillSetSkillAdminFormSet
    autocomplete_fields = ("eve_type",)

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related("eve_type", "skill_set__ship_type")

    def get_formset(self, *args, **kwargs):
        return super().get_formset(validate_min=True, *args, **kwargs)

    # def get_formset(self, *args, **kwargs):
    #     formset = super().get_formset(*args, **kwargs)
    #     qs = formset.form.base_fields["skill_set"].queryset
    #     qs = qs.select_related("skill_set__ship_type__eve_group")
    #     formset.form.base_fields["skill_set"].queryset = qs
    #     return formset


# class SkillSetShipTypeFilter(admin.SimpleListFilter):
#     title = "is ship type"
#     parameter_name = "is_ship_type"

#     def lookups(self, request, model_admin):
#         return (
#             ("yes", "yes"),
#             ("no", "no"),
#         )

#     def queryset(self, request, queryset):
#         if self.value() == "yes":
#             return SkillSet.objects.filter(ship_type__isnull=False)
#         if self.value() == "no":
#             return SkillSet.objects.filter(ship_type__isnull=True)
#         return SkillSet.objects.all()


@admin.register(SkillSet)
class SkillSetAdmin(AddDeleteObjects, admin.ModelAdmin):
    autocomplete_fields = ("ship_type",)
    list_display = (
        "name",
        "ship_type",
        "_skills",
        "_groups",
        "is_visible",
    )
    list_filter = (
        # SkillSetShipTypeFilter,  # this filter disables the prefetch in get_queryset
        ("groups", admin.RelatedOnlyFieldListFilter),
        "is_visible",
    )
    ordering = ["name"]
    search_fields = ["name"]
    actions = ["delete_objects", "clone_skill_sets"]

    fields = [
        "name",
        "description",
        "ship_type",
        "is_visible",
        ("last_modified_at", "last_modified_by"),
    ]
    readonly_fields = ("last_modified_at", "last_modified_by")
    inlines = (SkillSetSkillAdminInline,)

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related("ship_type").prefetch_related(
            Prefetch(
                "skills",
                queryset=SkillSetSkill.objects.select_related("eve_type").order_by(
                    "eve_type__name"
                ),
                to_attr="skills_ordered",
            ),
            Prefetch(
                "groups",
                queryset=SkillSetGroup.objects.order_by("name"),
                to_attr="groups_ordered",
            ),
        )

    def _skills(self, obj):
        skills = []
        for skill in obj.skills_ordered:
            skill_name = skill.eve_type.name
            required_level = skill.required_level if skill.required_level else ""
            recommended_level = (
                f"[{skill.recommended_level}]" if skill.recommended_level else ""
            )
            skills.append(f"{skill_name} {required_level} {recommended_level}")
        return skills

    def _groups(self, obj) -> Optional[List[str]]:
        groups = [f"{group.name}" for group in obj.groups_ordered]
        return groups if groups else None

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "ship_type":
            kwargs["queryset"] = (
                EveType.objects.select_related("eve_group__eve_category")
                .filter(eve_group__eve_category=EveCategoryId.SHIP)
                .order_by("name")
            )
        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    def save_model(self, request, obj, form, change):
        obj.user = request.user
        obj.last_modified_by = request.user
        obj.last_modified_at = now()
        super().save_model(request, obj, form, change)
        tasks.update_characters_skill_checks.apply_async(
            kwargs={"force_update": True}, priority=MEMBERAUDIT_TASKS_NORMAL_PRIORITY
        )  # type: ignore

    def delete_model(self, request, obj):
        obj.user = request.user
        super().delete_model(request, obj)
        tasks.update_characters_skill_checks.apply_async(
            kwargs={"force_update": True}, priority=MEMBERAUDIT_TASKS_NORMAL_PRIORITY
        )  # type: ignore

    @admin.action(description=_("Clone selected skill sets"))
    def clone_skill_sets(self, request, queryset):
        for obj in queryset:
            obj.clone(request.user)

        self.message_user(request, _("Cloned %d skill sets.") % queryset.count())



import datetime as dt
from unittest.mock import MagicMock, patch

from django.test import TestCase
from django.utils.timezone import now
from esi.errors import TokenError

from allianceauth.eveonline.models import EveCharacter
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.esi_testing import build_http_error
from app_utils.testing import NoSocketsTestCase, create_user_from_evecharacter

from memberaudit.helpers import UpdateSectionResult
from memberaudit.models import Character, CharacterUpdateStatus
from memberaudit.tests.testdata.factories import (
    create_character,
    create_character_from_user,
    create_character_update_status,
)
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.utils import (
    add_memberaudit_character_to_user,
    create_memberaudit_character,
    create_user_from_evecharacter_with_access,
)

MODULE_PATH = "memberaudit.models.characters"


class TestCharacterUserHasAccess(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.user_1001, _ = create_user_from_evecharacter_with_access(1001)

    def test_user_owning_character_has_access(self):
        # given
        character_1001 = create_character_from_user(self.user_1001)
        # when/then
        self.assertTrue(character_1001.user_has_access(self.user_1001))

    def test_other_user_has_no_access(self):
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_lex = AuthUtils.create_user("Lex_Luthor")
        # when/then
        self.assertFalse(character_1001.user_has_access(user_lex))

    def test_has_no_access_for_view_everything_without_scope_permission(self):
        # given
        character_1101 = create_memberaudit_character(1101)
        user_1002, _ = create_user_from_evecharacter(
            1002,
            permissions=["memberaudit.basic_access", "memberaudit.view_everything"],
        )
        # when/then
        self.assertFalse(character_1101.user_has_access(user_1002))

    def test_has_access_for_view_everything_with_scope_permission(self):
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1002, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "memberaudit.basic_access",
                "memberaudit.view_everything",
                "memberaudit.characters_access",
            ],
        )
        # when/then
        self.assertTrue(character_1001.user_has_access(user_1002))

    def test_has_access_for_view_everything_with_scope_permission_to_orphan(self):
        # given
        character_1121 = create_character(EveCharacter.objects.get(character_id=1121))
        user_1002, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "memberaudit.basic_access",
                "memberaudit.view_everything",
                "memberaudit.characters_access",
            ],
        )
        # when/then
        self.assertTrue(character_1121.user_has_access(user_1002))

    def test_view_same_corporation_1a(self):
        """
        when user has view_same_corporation permission and not characters_access
        and is in the same corporation as the character owner (main)
        then return False
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1002, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "memberaudit.basic_access",
                "memberaudit.view_same_corporation",
            ],
        )
        # when/then
        self.assertFalse(character_1001.user_has_access(user_1002))

    def test_view_same_corporation_1b(self):
        """
        when user has view_same_corporation permission and characters_access
        and is in the same corporation as the character owner (main)
        then return True
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1002, _ = create_user_from_evecharacter_with_access(1002)
        user_1002 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_corporation", user_1002
        )
        user_1002 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.characters_access", user_1002
        )
        # when/then
        self.assertTrue(character_1001.user_has_access(user_1002))

    def test_view_same_corporation_2a(self):
        """
        when user has view_same_corporation permission and not characters_access
        and is in the same corporation as the character owner (alt)
        then return False
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1002, _ = create_user_from_evecharacter_with_access(1002)
        user_1002 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_corporation", user_1002
        )
        character_1103 = add_memberaudit_character_to_user(
            character_1001.eve_character.character_ownership.user, 1103
        )
        # when/then
        self.assertFalse(character_1103.user_has_access(user_1002))

    def test_view_same_corporation_2b(self):
        """
        when user has view_same_corporation permission and characters_access
        and is in the same corporation as the character owner (alt)
        then return True
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1002, _ = create_user_from_evecharacter_with_access(1002)
        user_1002 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_corporation", user_1002
        )
        user_1002 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.characters_access", user_1002
        )
        character_1103 = add_memberaudit_character_to_user(
            character_1001.eve_character.character_ownership.user, 1103
        )
        self.assertTrue(character_1103.user_has_access(user_1002))

    def test_view_same_corporation_3(self):
        """
        when user has view_same_corporation permission and characters_access
        and is NOT in the same corporation as the character owner
        then return False
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1003, _ = create_user_from_evecharacter_with_access(1003)
        user_1003 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_corporation", user_1003
        )
        user_1003 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.characters_access", user_1003
        )
        # when/then
        self.assertFalse(character_1001.user_has_access(user_1003))

    def test_view_same_alliance_1a(self):
        """
        when user has view_same_alliance permission and not characters_access
        and is in the same alliance as the character's owner (main)
        then return False
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1003, _ = create_user_from_evecharacter_with_access(1003)
        user_1003 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_alliance", user_1003
        )
        # when/then
        self.assertFalse(character_1001.user_has_access(user_1003))

    def test_view_same_alliance_1b(self):
        """
        when user has view_same_alliance permission and characters_access
        and is in the same alliance as the character's owner (main)
        then return True
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1003, _ = create_user_from_evecharacter_with_access(1003)
        user_1003 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_alliance", user_1003
        )
        user_1003 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.characters_access", user_1003
        )
        # when/then
        self.assertTrue(character_1001.user_has_access(user_1003))

    def test_view_same_alliance_2a(self):
        """
        when user has view_same_alliance permission and not characters_access
        and is in the same alliance as the character's owner (alt)
        then return False
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1003, _ = create_user_from_evecharacter_with_access(1003)
        user_1003 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_alliance", user_1003
        )
        character_1103 = add_memberaudit_character_to_user(
            character_1001.eve_character.character_ownership.user, 1103
        )
        # when/then
        self.assertFalse(character_1103.user_has_access(user_1003))

    def test_view_same_alliance_2b(self):
        """
        when user has view_same_alliance permission and characters_access
        and is in the same alliance as the character's owner (alt)
        then return True
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1003, _ = create_user_from_evecharacter_with_access(1003)
        user_1003 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_alliance", user_1003
        )
        user_1003 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.characters_access", user_1003
        )
        character_1103 = add_memberaudit_character_to_user(
            character_1001.eve_character.character_ownership.user, 1103
        )
        # when/then
        self.assertTrue(character_1103.user_has_access(user_1003))

    def test_view_same_alliance_3(self):
        """
        when user has view_same_alliance permission and characters_access
        and is NOT in the same alliance as the character owner
        then return False
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1101, _ = create_user_from_evecharacter_with_access(1101)
        user_1101 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_alliance", user_1101
        )
        user_1101 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.characters_access", user_1101
        )
        # when/then
        self.assertFalse(character_1001.user_has_access(user_1101))

    def test_recruiter_access_1(self):
        """
        when user has recruiter permission
        and character is shared
        then return True
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        character_1001.is_shared = True
        character_1001.save()
        AuthUtils.add_permission_to_user_by_name(
            "memberaudit.share_characters",
            character_1001.eve_character.character_ownership.user,
        )
        user_1101, _ = create_user_from_evecharacter_with_access(1101)
        user_1101 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_shared_characters", user_1101
        )
        # when/then
        self.assertTrue(character_1001.user_has_access(user_1101))

    def test_recruiter_access_2(self):
        """
        when user has recruiter permission
        and character is NOT shared
        then return False
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        character_1001.is_shared = False
        character_1001.save()
        AuthUtils.add_permission_to_user_by_name(
            "memberaudit.share_characters",
            character_1001.eve_character.character_ownership.user,
        )
        user_1101, _ = create_user_from_evecharacter_with_access(1101)
        user_1101 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_shared_characters", user_1101
        )
        # when/then
        self.assertFalse(character_1001.user_has_access(user_1101))


class TestCharacterUserHasScope(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.user_1001, _ = create_user_from_evecharacter_with_access(1001)

    def test_user_owning_character_has_scope(self):
        # given
        character_1001 = create_character_from_user(self.user_1001)
        # when/then
        self.assertTrue(character_1001.user_has_scope(self.user_1001))

    def test_other_user_has_no_scope(self):
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_lex = AuthUtils.create_user("Lex Luthor")
        # when/then
        self.assertFalse(character_1001.user_has_scope(user_lex))

    def test_has_no_scope_for_view_everything_without_scope_permission(self):
        # given
        character_1001 = create_memberaudit_character(1101)
        user_1002, _ = create_user_from_evecharacter(1002)
        # when/then
        self.assertFalse(character_1001.user_has_scope(user_1002))

    def test_has_scope_for_view_everything_with_scope_permission(self):
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1002, _ = create_user_from_evecharacter(
            1002, permissions=["memberaudit.view_everything"]
        )
        # when/then
        self.assertTrue(character_1001.user_has_scope(user_1002))

    def test_has_scope_for_view_everything_with_scope_permission_to_orphan(self):
        # given
        character_1121 = create_character(EveCharacter.objects.get(character_id=1121))
        user_1002, _ = create_user_from_evecharacter(
            1002, permissions=["memberaudit.view_everything"]
        )
        # when/then
        self.assertTrue(character_1121.user_has_scope(user_1002))

    def test_view_same_corporation_1(self):
        """
        when user has view_same_corporation permission
        and is in the same corporation as the character owner (main)
        then return True
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1002, _ = create_user_from_evecharacter_with_access(1002)
        user_1002 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_corporation", user_1002
        )
        # when/then
        self.assertTrue(character_1001.user_has_scope(user_1002))

    def test_view_same_corporation_2(self):
        """
        when user has view_same_corporation permission
        and is in the same corporation as the character owner (alt)
        then return True
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1002, _ = create_user_from_evecharacter_with_access(1002)
        user_1002 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_corporation", user_1002
        )
        character_1103 = add_memberaudit_character_to_user(
            character_1001.eve_character.character_ownership.user, 1103
        )
        self.assertTrue(character_1103.user_has_scope(user_1002))

    def test_view_same_corporation_3(self):
        """
        when user has view_same_corporation permission
        and is NOT in the same corporation as the character owner
        then return False
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1003, _ = create_user_from_evecharacter_with_access(1003)
        user_1003 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_corporation", user_1003
        )
        # when/then
        self.assertFalse(character_1001.user_has_scope(user_1003))

    def test_view_same_alliance_1(self):
        """
        when user has view_same_alliance permission
        and is in the same alliance as the character's owner (main)
        then return True
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1003, _ = create_user_from_evecharacter_with_access(1003)
        user_1003 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_alliance", user_1003
        )
        # when/then
        self.assertTrue(character_1001.user_has_scope(user_1003))

    def test_view_same_alliance_2(self):
        """
        when user has view_same_alliance permission
        and is in the same alliance as the character's owner (alt)
        then return True
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1003, _ = create_user_from_evecharacter_with_access(1003)
        user_1003 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_alliance", user_1003
        )
        character_1103 = add_memberaudit_character_to_user(
            character_1001.eve_character.character_ownership.user, 1103
        )
        # when/then
        self.assertTrue(character_1103.user_has_scope(user_1003))

    def test_view_same_alliance_3(self):
        """
        when user has view_same_alliance permission
        and is NOT in the same alliance as the character owner
        then return False
        """
        # given
        character_1001 = create_character_from_user(self.user_1001)
        user_1101, _ = create_user_from_evecharacter_with_access(1101)
        user_1101 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_alliance", user_1101
        )
        # when/then
        self.assertFalse(character_1001.user_has_scope(user_1101))


@patch(MODULE_PATH + ".Character.update_section_content_hash")
@patch(MODULE_PATH + ".Character.has_section_changed")
class TestCharacterUpdateDataIfChangedOrForced(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character_1002 = create_memberaudit_character(1002)
        cls.user, _ = create_user_from_evecharacter_with_access(1001)

    @staticmethod
    def _fetch_func_template(character):
        return ["alpha"]

    @staticmethod
    def _store_func_template(character, data):
        pass

    def test_should_store_data_when_changed(
        self, mock_has_section_changed, mock_update_section_content_hash
    ):
        # given
        character = create_character_from_user(self.user)
        fetch_func_mock = MagicMock(side_effect=self._fetch_func_template)
        store_func_mock = MagicMock(side_effect=self._store_func_template)
        mock_has_section_changed.return_value = True
        # when
        result = character.update_section_if_changed(
            section=character.UpdateSection.LOCATION,
            fetch_func=fetch_func_mock,
            store_func=store_func_mock,
            force_update=False,
        )
        # then
        self.assertTrue(fetch_func_mock.called)
        self.assertTrue(store_func_mock.called)
        args, _ = store_func_mock.call_args
        self.assertEqual(args[1], ["alpha"])
        self.assertTrue(mock_update_section_content_hash.called)
        _, kwargs = mock_update_section_content_hash.call_args
        self.assertEqual(kwargs["content"], ["alpha"])
        self.assertListEqual(result.data, ["alpha"])
        self.assertTrue(result.is_changed)
        self.assertTrue(result.is_updated)

    def test_should_not_store_data_when_not_changed(
        self, mock_has_section_changed, mock_update_section_content_hash
    ):
        # given
        character = create_character_from_user(self.user)
        fetch_func_mock = MagicMock(side_effect=self._fetch_func_template)
        store_func_mock = MagicMock(side_effect=self._store_func_template)
        mock_has_section_changed.return_value = False
        # when
        result = character.update_section_if_changed(
            section=character.UpdateSection.LOCATION,
            fetch_func=fetch_func_mock,
            store_func=store_func_mock,
            force_update=False,
        )
        # then
        self.assertTrue(fetch_func_mock.called)
        self.assertFalse(store_func_mock.called)
        self.assertFalse(mock_update_section_content_hash.called)
        self.assertIsNone(result.data)
        self.assertFalse(result.is_changed)
        self.assertFalse(result.is_updated)

    def test_should_always_store_data_when_forced(
        self, mock_has_section_changed, mock_update_section_content_hash
    ):
        # given
        character = create_character_from_user(self.user)
        fetch_func_mock = MagicMock(side_effect=self._fetch_func_template)
        store_func_mock = MagicMock(side_effect=self._store_func_template)
        mock_has_section_changed.return_value = False
        # when
        result = character.update_section_if_changed(
            section=character.UpdateSection.LOCATION,
            fetch_func=fetch_func_mock,
            store_func=store_func_mock,
            force_update=True,
        )
        # then
        self.assertTrue(fetch_func_mock.called)
        self.assertTrue(store_func_mock.called)
        self.assertTrue(mock_update_section_content_hash.called)
        self.assertListEqual(result.data, ["alpha"])
        self.assertFalse(result.is_changed)
        self.assertTrue(result.is_updated)

    def test_should_not_store_anything_when_esi_returns_http_500_and_return_none(
        self, mock_has_section_changed, mock_update_section_content_hash
    ):
        # given
        character = create_character_from_user(self.user)
        fetch_func_mock = MagicMock(side_effect=build_http_error(500, "Test exception"))
        store_func_mock = MagicMock(side_effect=self._store_func_template)
        mock_has_section_changed.side_effect = RuntimeError("Should not be called")
        # when
        result = character.update_section_if_changed(
            section=character.UpdateSection.LOCATION,
            fetch_func=fetch_func_mock,
            store_func=store_func_mock,
            force_update=False,
        )
        # then
        self.assertTrue(fetch_func_mock.called)
        self.assertFalse(store_func_mock.called)
        self.assertFalse(mock_update_section_content_hash.called)
        self.assertIsNone(result.is_changed)
        self.assertFalse(result.is_updated)

    def test_should_store_data_when_changed_and_use_hash_num(
        self, mock_has_section_changed, mock_update_section_content_hash
    ):
        # given
        character = create_character_from_user(self.user)
        fetch_func_mock = MagicMock(side_effect=self._fetch_func_template)
        store_func_mock = MagicMock(side_effect=self._store_func_template)
        mock_has_section_changed.return_value = True
        # when
        character.update_section_if_changed(
            section=character.UpdateSection.LOCATION,
            fetch_func=fetch_func_mock,
            store_func=store_func_mock,
            force_update=False,
            hash_num=2,
        )
        # then
        self.assertTrue(fetch_func_mock.called)
        self.assertTrue(store_func_mock.called)
        args, _ = store_func_mock.call_args
        self.assertEqual(args[1], ["alpha"])
        _, kwargs = mock_has_section_changed.call_args
        self.assertEqual(kwargs["hash_num"], 2)
        _, kwargs = mock_update_section_content_hash.call_args
        self.assertEqual(kwargs["hash_num"], 2)

    def test_should_skip_storing_data_when_no_store_func_provided(
        self, mock_has_section_changed, mock_update_section_content_hash
    ):
        # given
        character = create_character_from_user(self.user)
        fetch_func_mock = MagicMock(side_effect=self._fetch_func_template)
        mock_has_section_changed.return_value = True
        # when
        result = character.update_section_if_changed(
            section=character.UpdateSection.LOCATION,
            fetch_func=fetch_func_mock,
            store_func=None,
        )
        # then
        self.assertTrue(fetch_func_mock.called)
        self.assertTrue(mock_update_section_content_hash.called)
        self.assertListEqual(result.data, ["alpha"])
        self.assertTrue(result.is_changed)
        self.assertFalse(result.is_updated)

    @patch(MODULE_PATH + ".EveEntity.objects.bulk_resolve_ids")
    def test_should_resolve_eve_entity_ids_when_provided(
        self,
        mock_bulk_resolve_ids,
        mock_has_section_changed,
        mock_update_section_content_hash,
    ):
        # given
        def my_store_func(character, data):
            return [1, 2]

        fetch_func_mock = MagicMock(side_effect=self._fetch_func_template)
        mock_has_section_changed.return_value = True
        # when
        self.character_1002.update_section_if_changed(
            section=Character.UpdateSection.LOCATION,
            fetch_func=fetch_func_mock,
            store_func=my_store_func,
            force_update=False,
        )
        # then
        self.assertTrue(fetch_func_mock.called)
        self.assertTrue(mock_bulk_resolve_ids.called)
        args, _ = mock_bulk_resolve_ids.call_args
        self.assertListEqual(args[0], [1, 2])

    @patch(MODULE_PATH + ".EveEntity.objects.bulk_resolve_ids")
    def test_should_not_resolve_eve_entity_ids_when_not_provided(
        self,
        mock_bulk_resolve_ids,
        mock_has_section_changed,
        mock_update_section_content_hash,
    ):
        # given
        def my_store_func(character, data):
            return []

        fetch_func_mock = MagicMock(side_effect=self._fetch_func_template)
        mock_has_section_changed.return_value = True
        # when
        self.character_1002.update_section_if_changed(
            section=Character.UpdateSection.LOCATION,
            fetch_func=fetch_func_mock,
            store_func=my_store_func,
            force_update=False,
        )
        # then
        self.assertTrue(fetch_func_mock.called)
        self.assertFalse(mock_bulk_resolve_ids.called)


class TestCharacterHasTokenError(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character = create_memberaudit_character(1001)

    def test_should_return_false_when_no_error(self):
        # when/then
        self.assertFalse(self.character.has_token_issue())

    def test_should_return_true_when_token_error(self):
        # given
        create_character_update_status(
            self.character,
            section=Character.UpdateSection.ASSETS,
            is_success=False,
            has_token_error=True,
            error_message="TokenError",
        )
        # when/then
        self.assertTrue(self.character.has_token_issue())

    def test_should_return_false_when_other_error(self):
        # given
        create_character_update_status(
            self.character,
            section=Character.UpdateSection.ASSETS,
            is_success=False,
            has_token_error=False,
            error_message="other error",
        )
        # when/then
        self.assertFalse(self.character.has_token_issue())

    @patch(MODULE_PATH + ".MEMBERAUDIT_FEATURE_ROLES_ENABLED", False)
    def test_should_return_false_when_token_error_for_disabled_section(self):
        # given
        create_character_update_status(
            self.character,
            section=Character.UpdateSection.ROLES,
            is_success=False,
            has_token_error=True,
            error_message="TokenError",
        )
        # when/then
        self.assertFalse(self.character.has_token_issue())


class TestCharacterResetTokenErrorNotifiedIfStatusOk(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.user, _ = create_user_from_evecharacter_with_access(1001)

    def test_should_reset_when_ok_again(self):
        # given
        character = create_character_from_user(self.user, token_error_notified_at=now())
        for section in Character.UpdateSection:
            create_character_update_status(character, section=section)

        # when
        character.reset_token_error_notified_if_status_ok()

        # then
        character.refresh_from_db()
        self.assertIsNone(character.token_error_notified_at)

    def test_should_not_reset_when_not_yet_ok(self):
        # given
        character = create_character_from_user(self.user, token_error_notified_at=now())
        create_character_update_status(
            character, section=Character.UpdateSection.ASSETS, is_success=False
        )

        # when
        character.reset_token_error_notified_if_status_ok()

        # then
        character.refresh_from_db()
        self.assertTrue(character.token_error_notified_at)

    def test_should_ignore_when_not_set(self):
        # given
        character = create_character_from_user(self.user, token_error_notified_at=None)
        create_character_update_status(
            character, section=Character.UpdateSection.ASSETS, is_success=False
        )

        # when
        character.reset_token_error_notified_if_status_ok()

        # then
        character.refresh_from_db()
        self.assertIsNone(character.token_error_notified_at)


class TestCharacterGetEsiScopes(TestCase):
    @patch(MODULE_PATH + ".MEMBERAUDIT_FEATURE_ROLES_ENABLED", False)
    def test_should_return_all_scopes(self):
        # when
        result = Character.get_esi_scopes()
        # then
        expected = {
            "esi-assets.read_assets.v1",
            "esi-bookmarks.read_character_bookmarks.v1",
            "esi-calendar.read_calendar_events.v1",
            "esi-characters.read_agents_research.v1",
            "esi-characters.read_blueprints.v1",
            "esi-characters.read_contacts.v1",
            "esi-characters.read_corporation_roles.v1",  # NEW
            "esi-characters.read_fatigue.v1",
            "esi-characters.read_fw_stats.v1",
            "esi-characters.read_loyalty.v1",
            "esi-characters.read_medals.v1",
            "esi-characters.read_notifications.v1",
            "esi-characters.read_opportunities.v1",
            "esi-characters.read_standings.v1",
            "esi-characters.read_titles.v1",
            "esi-clones.read_clones.v1",
            "esi-clones.read_implants.v1",
            "esi-contracts.read_character_contracts.v1",
            "esi-corporations.read_corporation_membership.v1",
            "esi-industry.read_character_jobs.v1",
            "esi-industry.read_character_mining.v1",
            "esi-killmails.read_killmails.v1",
            "esi-location.read_location.v1",
            "esi-location.read_online.v1",
            "esi-location.read_ship_type.v1",
            "esi-mail.read_mail.v1",
            "esi-markets.read_character_orders.v1",
            "esi-markets.structure_markets.v1",
            "esi-planets.manage_planets.v1",
            "esi-planets.read_customs_offices.v1",
            "esi-search.search_structures.v1",
            "esi-skills.read_skillqueue.v1",
            "esi-skills.read_skills.v1",
            "esi-universe.read_structures.v1",
            "esi-wallet.read_character_wallet.v1",
        }
        self.assertSetEqual(set(result), expected)


class TestCharacterPerformUpdateWithErrorLogging(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character = create_memberaudit_character(1001)

    def setUp(self) -> None:
        self.character.update_status_set.all().delete()

    def test_should_execute_method_and_return_value(self):
        # given
        def my_method(dummy):
            return UpdateSectionResult(
                data=f"return-value-{dummy}", is_changed=True, is_updated=True
            )

        section = Character.UpdateSection.LOCATION
        # when
        result = self.character.perform_update_with_error_logging(
            section=section, method=my_method, dummy="alpha"
        )
        # then
        self.assertEqual(result.data, "return-value-alpha")
        self.assertTrue(result.is_updated)

    def test_should_mark_section_as_failed_when_general_exception_is_raised(self):
        # given
        def my_method():
            raise RuntimeError("Test exception")

        section = Character.UpdateSection.LOCATION
        # when/then
        with self.assertRaises(RuntimeError):
            self.character.perform_update_with_error_logging(
                section=section, method=my_method
            )
        # then
        status: CharacterUpdateStatus = self.character.update_status_set.get(
            section=section
        )
        self.assertFalse(status.is_success)
        self.assertFalse(status.has_token_error)
        self.assertIn("RuntimeError", status.error_message)
        self.assertTrue(status.run_finished_at)

    def test_should_mark_section_as_failed_when_token_error_is_raised(self):
        # given
        def my_method():
            raise TokenError("Test exception")

        section = Character.UpdateSection.LOCATION
        # when/then
        with self.assertRaises(TokenError):
            self.character.perform_update_with_error_logging(
                section=section, method=my_method
            )
        # then
        status: CharacterUpdateStatus = self.character.update_status_set.get(
            section=section
        )
        self.assertFalse(status.is_success)
        self.assertTrue(status.has_token_error)
        self.assertIn("TokenError", status.error_message)
        self.assertTrue(status.run_finished_at)


class TestCharacterUpdateStatusAsDict(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character = create_memberaudit_character(1001)

    def test_should_return_dict_with_status(self):
        # given
        status = create_character_update_status(
            self.character, section=Character.UpdateSection.LOCATION, is_success=True
        )
        # when
        result = self.character.update_status_as_dict()
        # then
        self.assertDictEqual(result, {"location": status})

    def test_should_return_empty_dict(self):
        # when
        result = self.character.update_status_as_dict()
        # then
        self.assertDictEqual(result, {})


@patch(MODULE_PATH + ".section_time_until_stale", {"assets": 640})
class TestCharacterUpdateStatusIsUpdateNeeded(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.section = Character.UpdateSection.ASSETS
        cls.character = create_memberaudit_character(1001)

    def test_should_report_false_when_section_not_stale(self):
        # given
        status = create_character_update_status(
            character=self.character,
            section=self.section,
            is_success=True,
            run_started_at=now() - dt.timedelta(seconds=30),
            run_finished_at=now(),
        )
        # when/then
        self.assertFalse(status.is_update_needed())

    def test_should_report_true_when_section_has_error(self):
        # given
        status = create_character_update_status(
            character=self.character, section=self.section, is_success=False
        )
        # when/then
        self.assertTrue(status.is_update_needed())

    def test_should_report_true_when_section_is_stale(self):
        # given
        run_started_at = now() - dt.timedelta(hours=12)
        run_finished_at = run_started_at + dt.timedelta(minutes=10)
        status = create_character_update_status(
            character=self.character,
            section=self.section,
            is_success=True,
            run_started_at=run_started_at,
            run_finished_at=run_finished_at,
        )
        # when/then
        self.assertTrue(status.is_update_needed())

    def test_should_report_false_when_section_has_token_error_and_stale(self):
        # given
        run_started_at = now() - dt.timedelta(hours=12)
        status = create_character_update_status(
            character=self.character,
            section=self.section,
            is_success=False,
            run_started_at=run_started_at,
            has_token_error=True,
        )
        # when/then
        self.assertFalse(status.is_update_needed())

    def test_should_report_false_when_section_has_token_error_and_not_stale(self):
        # given
        status = create_character_update_status(
            character=self.character,
            section=self.section,
            is_success=False,
            has_token_error=True,
        )
        # when/then
        self.assertFalse(status.is_update_needed())



"""Logic shared by managers."""

# pylint: disable=too-many-positional-arguments

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Sequence, Set

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from memberaudit import __title__
from memberaudit.app_settings import MEMBERAUDIT_BULK_METHODS_BATCH_SIZE
from memberaudit.helpers import eve_entity_ids_from_objs

if TYPE_CHECKING:
    from memberaudit.models import Character


logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class GenericUpdateSimpleObjMixin:
    """Adds the ability to update objs from ESI data.

    This is a generic implementation that works for any section,
    which data can be represented as a list of key/value pairs.
    """

    def _update_or_create_objs_generic(
        self,
        character: Character,
        esi_data: List[Dict[str, Any]],
        esi_fields: Sequence[str],
        model_fields: Sequence[str],
        make_obj_from_esi_entry: Callable,
        return_new_eve_entities: bool = False,
    ) -> Set[int]:
        """Update or create objs from esi data.

        Optionally returns EveEntity IDs from new objs,
        which can then be resolved in bulk.
        """
        if not esi_data:
            self.filter(character=character).delete()
            logger.info("%s: No %s", character, self.model._meta.verbose_name_plural)
            return set()

        current_entries = {obj[0]: obj[1] for obj in self.values_list(*model_fields)}
        incoming_entries = {obj[esi_fields[0]]: obj[esi_fields[1]] for obj in esi_data}

        new_entries_keys = self._create_new_objs(
            character, current_entries, incoming_entries, make_obj_from_esi_entry
        )
        self._update_modified_objs(
            character,
            current_entries,
            incoming_entries,
            key_field=model_fields[0],
            value_field=model_fields[1],
        )
        self._delete_obsolete_objs(
            character, current_entries, incoming_entries, key_field=model_fields[0]
        )

        if return_new_eve_entities:
            return new_entries_keys

        return set()

    def _create_new_objs(
        self,
        character: Character,
        current_entries: dict,
        incoming_entries: dict,
        make_obj_from_esi_entry: Callable,
    ) -> Set[int]:
        """Create new objects. Returns keys from new objects."""
        new_entries = {
            entity_id: standing
            for entity_id, standing in incoming_entries.items()
            if entity_id not in current_entries
        }
        if not new_entries:
            return set()

        objs = [
            make_obj_from_esi_entry(character, key, value)
            for key, value in new_entries.items()
        ]
        self.bulk_create(objs, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE)
        logger.info(
            "%s: Created %d new %s",
            character,
            len(objs),
            self.model._meta.verbose_name_plural,
        )
        return set(new_entries.keys())

    def _update_modified_objs(
        self,
        character: Character,
        current_entries: dict,
        incoming_entries: dict,
        key_field: str,
        value_field: str,
    ) -> None:
        modified_entries = {
            key: value
            for key, value in incoming_entries.items()
            if key in current_entries and current_entries[key] != value
        }
        if not modified_entries:
            return

        params = {"character": character, f"{key_field}__in": modified_entries}
        objs = self.filter(**params).in_bulk()
        for obj in objs.values():
            key = getattr(obj, key_field)
            setattr(obj, value_field, modified_entries[key])

        self.bulk_update(
            objs=objs.values(),
            fields=[value_field],
            batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE,
        )
        logger.info(
            "%s: Updated %d %s",
            character,
            len(objs),
            self.model._meta.verbose_name_plural,
        )

    def _delete_obsolete_objs(
        self,
        character: Character,
        current_entries: dict,
        incoming_entries: dict,
        key_field: str,
    ) -> None:
        obsolete_entries = {
            key: value
            for key, value in current_entries.items()
            if key not in incoming_entries
        }
        if not obsolete_entries:
            return

        params = {"character": character, f"{key_field}__in": obsolete_entries}
        self.filter(**params).delete()
        logger.info(
            "%s: Removed %d obsolete %s",
            character,
            len(obsolete_entries),
            self.model._meta.verbose_name_plural,
        )


class GenericUpdateComplexObjMixin:
    """Adds the ability to update objs from ESI data.

    This is a generic implementation that works for any section,
    which data has a functional primary key.
    """

    def _update_or_create_objs_generic(
        self,
        character: Character,
        esi_data: List[Dict[str, Any]],
        model_key_field: str,
        fields_for_update: Iterable[str],
        make_obj_from_esi_entry: Callable,
        return_new_eve_entities: bool = False,
    ) -> Set[int]:
        """Update or create objs from esi data."""
        if not esi_data:
            self.filter(character=character).delete()
            logger.info("%s: No %s", character, self.model._meta.verbose_name_plural)
            return set()

        current_objs = {
            getattr(obj, model_key_field): obj
            for obj in self.filter(character=character).in_bulk().values()
        }
        incoming_objs = {
            getattr(obj, model_key_field): obj
            for obj in [make_obj_from_esi_entry(character, entry) for entry in esi_data]
        }

        new_eve_entity_ids = self._create_new_objs(
            character, current_objs, incoming_objs, return_new_eve_entities
        )
        self._update_modified_objs(
            character, current_objs, incoming_objs, fields_for_update
        )
        self._delete_obsolete_objs(
            character, current_objs, incoming_objs, key_field=model_key_field
        )

        if return_new_eve_entities:
            return new_eve_entity_ids

        return set()

    def _create_new_objs(
        self,
        character: Character,
        current_objs: dict,
        incoming_objs: dict,
        return_new_eve_entities: bool,
    ) -> Set[int]:
        new_objs = [
            obj for key, obj in incoming_objs.items() if key not in current_objs
        ]
        if not new_objs:
            return set()

        self.bulk_create(new_objs, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE)
        logger.info(
            "%s: Created %d new %s",
            character,
            len(new_objs),
            self.model._meta.verbose_name_plural,
        )

        if return_new_eve_entities:
            return eve_entity_ids_from_objs(new_objs)

        return set()

    def _update_modified_objs(
        self,
        character: Character,
        current_objs: dict,
        incoming_objs: dict,
        fields_for_update: List[str],
    ) -> None:
        modified_objs = []
        for key, incoming_obj in incoming_objs.items():
            if key not in current_objs:
                continue

            current_obj = current_objs[key]
            has_changed = False
            for field in fields_for_update:
                new_value = getattr(incoming_obj, field)
                if getattr(current_obj, field) != new_value:
                    setattr(current_obj, field, new_value)
                    has_changed = True

            if has_changed:
                modified_objs.append(current_obj)

        if not modified_objs:
            return

        self.bulk_update(
            objs=modified_objs,
            fields=fields_for_update,
            batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE,
        )
        logger.info(
            "%s: Updated %d %s",
            character,
            len(modified_objs),
            self.model._meta.verbose_name_plural,
        )

    def _delete_obsolete_objs(
        self,
        character: Character,
        current_objs: dict,
        incoming_objs: dict,
        key_field: str,
    ) -> None:
        obsolete_obj_ids = {key for key in current_objs if key not in incoming_objs}
        if not obsolete_obj_ids:
            return

        params = {"character": character, f"{key_field}__in": obsolete_obj_ids}
        self.filter(**params).delete()
        logger.info(
            "%s: Removed %d obsolete %s",
            character,
            len(obsolete_obj_ids),
            self.model._meta.verbose_name_plural,
        )



from enum import IntEnum


class EvePlanetIds(IntEnum):
    AMAMAKE_I = 40161463
    AMAMAKE_II = 40161464
    AMAMAKE_III = 40161467
    AMAMAKE_IV = 40161469


class EveSolarSystemIds(IntEnum):
    AMAMAKE = 30002537



from typing import NamedTuple
from unittest import TestCase

from memberaudit.core.standings import Standing, calc_effective_standing


class TestCharacterContactStandingLevel(TestCase):
    class MyTestCase(NamedTuple):
        standing: float
        expected_result: str

    def test_should_determine_correct_standing(self):
        # given
        test_cases = [
            self.MyTestCase(9.9, Standing.EXCELLENT),
            self.MyTestCase(4.9, Standing.GOOD),
            self.MyTestCase(0.0, Standing.NEUTRAL),
            self.MyTestCase(-4.9, Standing.BAD),
            self.MyTestCase(-9.9, Standing.TERRIBLE),
        ]
        for test_case in test_cases:
            with self.subTest(standing=test_case.standing):
                # when
                standing = Standing.from_value(test_case.standing)
                # then
                self.assertEqual(standing, test_case.expected_result)


class TestCalcEffectiveStanding(TestCase):
    def test_should_calc_correct_standing(self):
        # given
        class MyTestCase(NamedTuple):
            unadjusted_standing: float
            skill_level: int
            skill_modifier: float
            max_possible_standing: int
            effective_standing: float

        test_cases = [
            MyTestCase(
                unadjusted_standing=0.9,
                skill_level=2,
                skill_modifier=0.04,
                max_possible_standing=10,
                effective_standing=1.628,
            ),
            MyTestCase(
                unadjusted_standing=7.49,
                skill_level=5,
                skill_modifier=0.04,
                max_possible_standing=10,
                effective_standing=7.99,
            ),
            MyTestCase(
                unadjusted_standing=-4.76,
                skill_level=5,
                skill_modifier=0.04,
                max_possible_standing=10,
                effective_standing=-1.81,
            ),
            MyTestCase(
                unadjusted_standing=-2.29,
                skill_level=5,
                skill_modifier=0.04,
                max_possible_standing=10,
                effective_standing=0.17,
            ),
            MyTestCase(
                unadjusted_standing=4.99,
                skill_level=3,
                skill_modifier=0.04,
                max_possible_standing=10,
                effective_standing=5.59,
            ),
        ]
        for num, test_case in enumerate(test_cases):
            with self.subTest(num=num):
                # when
                result = calc_effective_standing(
                    test_case.unadjusted_standing,
                    test_case.skill_level,
                    test_case.skill_modifier,
                    test_case.max_possible_standing,
                )
                # then
                self.assertAlmostEqual(result, test_case.effective_standing, 2)



"""Start update for updateable characters."""

from tqdm import tqdm

from django.core.management.base import BaseCommand, CommandError

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from memberaudit import __title__, tasks
from memberaudit.constants import IS_TESTING
from memberaudit.models import Character

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class Command(BaseCommand):
    help = str(__doc__)

    def add_arguments(self, parser):
        parser.add_argument("sections", nargs="*", help="sections to update")
        parser.add_argument(
            "--all",
            action="store_true",
            help="Update all sections",
        )

        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_true",
            help="Do NOT prompt the user for input of any kind",
        )

    def handle(self, *args, **options):
        updateable_characters = Character.objects.filter(
            is_disabled=False,
            eve_character__character_ownership__isnull=False,
        )

        if options["all"]:
            selected_sections = Character.UpdateSection.enabled_sections()

        elif options["sections"]:
            selected_sections = []
            for section_name in options["sections"]:
                try:
                    selected_sections.append(Character.UpdateSection(section_name))
                except ValueError:
                    raise CommandError(
                        f"'{section_name}' is not a valid section"
                    ) from None

        else:
            raise CommandError("You must specify which sections to update.")

        sections_text = ", ".join(
            sorted(str(section.label) for section in selected_sections)
        )
        updatable_count = updateable_characters.count()
        all_count = Character.objects.count()
        self.stdout.write(
            f"{updatable_count} of {all_count} character are currently updateable "
            "(i.e. are not disabled or orphans)."
        )
        self.stdout.write(
            "Are you sure you want to start updating the following sections "
            "for these characters: "
        )
        self.stdout.write(f"{sections_text}? ", ending="")
        answer = input("[Y/n]") if not options["noinput"] else "y"
        if answer.lower() == "n":
            self.stdout.write(self.style.WARNING("Aborted"))
            return

        for character in tqdm(
            updateable_characters,
            desc="Starting update tasks",
            total=updatable_count,
            leave=False,
            disable=IS_TESTING,
            unit="character",
        ):
            for section in selected_sections:
                task_name = f"update_character_{section.value}"
                task = getattr(tasks, task_name)
                task.apply_async(
                    kwargs={"character_pk": character.pk, "force_update": True},
                    priority=tasks.MEMBERAUDIT_TASKS_LOW_PRIORITY,
                )

        msg = (
            f"Started forced update of {sections_text} for {updatable_count} characters"
        )
        logger.info(msg)
        self.stdout.write(msg)
        self.stdout.write(self.style.SUCCESS("Done"))



"""Member Audit forms."""

from django import forms
from django.core.exceptions import ValidationError

from memberaudit.core.eft_parser import EftParserError, create_fitting_from_eft
from memberaudit.core.skill_plans import SkillPlan, SkillPlanError
from memberaudit.models import SkillSet, SkillSetGroup
from memberaudit.models.constants import NAMES_MAX_LENGTH


class ImportFittingForm(forms.Form):
    """Form for importing an EFT fitting."""

    fitting_text = forms.CharField(
        label="",
        widget=forms.Textarea(
            attrs={
                "placeholder": "Paste fitting in EFT format into this field...",
                "rows": 30,
                "cols": 100,
            }
        ),
    )
    can_overwrite = forms.BooleanField(
        label="Overwrite skill sets with same name", required=False
    )
    skill_set_name = forms.CharField(
        label="New Name",
        max_length=NAMES_MAX_LENGTH,
        required=False,
        widget=forms.TextInput(attrs={"placeholder": "Instead of name from fitting."}),
    )
    skill_set_group = forms.ModelChoiceField(
        label="Add to skill set group",
        required=False,
        queryset=SkillSetGroup.objects.order_by("name"),
    )

    def clean(self):
        data = super().clean()
        try:
            fitting, errors = create_fitting_from_eft(data["fitting_text"])
        except EftParserError:
            raise ValidationError(
                {
                    "fitting_text": (
                        "This fitting does not appear to be a valid EFT format."
                    )
                }
            ) from None

        data["_fitting"] = fitting
        data["_errors"] = errors
        skill_set_name = data.get("skill_set_name") or fitting.name
        if (
            not data["can_overwrite"]
            and SkillSet.objects.filter(name=skill_set_name).exists()
        ):
            raise ValidationError(
                f"A skill set with the name '{skill_set_name}' already exists."
            )
        data["_skill_set_name"] = skill_set_name
        return data


class ImportSkillPlanForm(forms.Form):
    """Form for importing a skill plan."""

    skill_plan_text = forms.CharField(
        label="",
        widget=forms.Textarea(
            attrs={
                "placeholder": "Paste skill plan from Eve Online or Eve Mon into this field...",
                "rows": 30,
                "cols": 100,
            }
        ),
    )
    skill_set_name = forms.CharField(
        label="Name",
        max_length=NAMES_MAX_LENGTH,
        widget=forms.TextInput(attrs={"placeholder": "Name of the new skill set"}),
    )
    can_overwrite = forms.BooleanField(
        label="Overwrite skill sets with same name", required=False
    )
    skill_set_group = forms.ModelChoiceField(
        label="Add to skill set group",
        required=False,
        queryset=SkillSetGroup.objects.order_by("name"),
    )

    def clean(self):
        data = super().clean()
        try:
            skill_plan, errors = SkillPlan.create_from_plain_text(
                name=data["skill_set_name"], text=data["skill_plan_text"]
            )
        except SkillPlanError:
            raise ValidationError(
                {"skill_plan_text": "This does not appear to be a valid skill plan."}
            ) from None

        data["_skill_plan"] = skill_plan
        data["_errors"] = errors
        if (
            not data["can_overwrite"]
            and SkillSet.objects.filter(name=data["skill_set_name"]).exists()
        ):
            raise ValidationError(
                {"skill_set_name": "A skill set with this name already exists."}
            )
        return data



from eveuniverse.models import EveType

from app_utils.testing import NoSocketsTestCase

from memberaudit import forms
from memberaudit.core.fittings import Fitting

from .testdata.factories import (
    create_fitting_text,
    create_skill,
    create_skill_plan,
    create_skill_set,
)
from .testdata.load_eveuniverse import load_eveuniverse


class TestImportFittingForm(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.fitting_text = create_fitting_text("fitting_tristan.txt")

    def test_should_raise_no_errors(self):
        # given
        form = forms.ImportFittingForm(data={"fitting_text": self.fitting_text})
        # when
        self.assertTrue(form.is_valid())
        self.assertIsInstance(form.cleaned_data["_fitting"], Fitting)
        self.assertListEqual(form.cleaned_data["_errors"], [])

    def test_should_raise_error_when_fitting_text_is_invalid(self):
        # given
        form = forms.ImportFittingForm(data={"fitting_text": "invalid"})
        # when
        self.assertFalse(form.is_valid())
        self.assertIn("fitting_text", form.errors.keys())

    def test_should_raise_error_when_skill_set_already_exists_1(self):
        # given
        create_skill_set(name="Tristan - Standard Kite (cap stable)")
        form = forms.ImportFittingForm(data={"fitting_text": self.fitting_text})
        # when
        self.assertFalse(form.is_valid())

    def test_should_raise_error_when_skill_set_already_exists_2(self):
        # given
        create_skill_set(name="My name")
        form = forms.ImportFittingForm(
            data={"fitting_text": self.fitting_text, "skill_set_name": "My name"}
        )
        # when
        self.assertFalse(form.is_valid())


class TestImportSkillPlanForm(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_should_raise_no_errors(self):
        # given
        form = forms.ImportSkillPlanForm(
            data={
                "skill_plan_text": "Caldari Core Systems 5",
                "skill_set_name": "My name",
            }
        )
        # when
        self.assertTrue(form.is_valid())
        expected_skill_plan = create_skill_plan(
            name="My name",
            skills=[
                create_skill(
                    eve_type=EveType.objects.get(name="Caldari Core Systems"), level=5
                )
            ],
        )
        self.assertEqual(form.cleaned_data["_skill_plan"], expected_skill_plan)
        self.assertListEqual(form.cleaned_data["_errors"], [])

    def test_should_raise_error_when_skill_set_already_exists(self):
        # given
        create_skill_set(name="My name")
        form = forms.ImportSkillPlanForm(
            data={
                "skill_plan_text": "Caldari Core Systems 5",
                "skill_set_name": "My name",
            }
        )
        # when
        self.assertFalse(form.is_valid())
        self.assertIn("skill_set_name", form.errors.keys())

    def test_should_raise_error_when_skill_plan_invalid(self):
        # given
        create_skill_set(name="My name")
        form = forms.ImportSkillPlanForm(
            data={
                "skill_plan_text": "Invalid skill plan",
                "skill_set_name": "My name",
            }
        )
        # when
        self.assertFalse(form.is_valid())
        self.assertIn("skill_plan_text", form.errors.keys())



# from django.test import TestCase

from eveuniverse.models import EveType

from app_utils.testing import NoSocketsTestCase

from memberaudit.core.fittings import Fitting, Module, _BaseFittingItem
from memberaudit.tests.testdata.factories import create_fitting
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse


class TestBaseFittingItem(NoSocketsTestCase):
    def test_can_not_instantiate(self):
        # given
        with self.assertRaises(RuntimeError):
            _BaseFittingItem()


class TestModule(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_should_be_empty(self):
        # given
        module = Module()
        # when/then
        self.assertTrue(module.is_empty)

    def test_should_not_be_empty(self):
        # given
        drones = EveType.objects.get(name="Drone Damage Amplifier II")
        module = Module(module_type=drones)
        # when/then
        self.assertFalse(module.is_empty)

    def test_can_not_convert_empty_module_to_eft(self):
        # given
        module = Module()
        # when/then
        with self.assertRaises(ValueError):
            module.to_eft()

    def test_can_convert_simple_module_to_eft(self):
        # given
        module_type = EveType.objects.get(name="Drone Damage Amplifier II")
        module = Module(module_type=module_type)
        # when
        result = module.to_eft()
        # then
        self.assertEqual(result, "Drone Damage Amplifier II")

    def test_can_convert_module_with_charge_to_eft(self):
        # given
        module_type = EveType.objects.get(name="125mm Gatling AutoCannon II")
        charge_type = EveType.objects.get(name="EMP S")
        module = Module(module_type=module_type, charge_type=charge_type)
        # when
        result = module.to_eft()
        # then
        self.assertEqual(result, "125mm Gatling AutoCannon II, EMP S")

    def test_can_convert_offline_module_to_eft(self):
        # given
        module_type = EveType.objects.get(name="Drone Damage Amplifier II")
        module = Module(module_type=module_type, is_offline=True)
        # when
        result = module.to_eft()
        # then
        self.assertEqual(result, "Drone Damage Amplifier II /OFFLINE")


class TestFitting(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_should_return_str(self):
        # given
        ship_type = EveType.objects.get(name="Svipul")
        fitting = Fitting("Dummy Fitting", ship_type)
        # when/then
        self.assertEqual(str(fitting), "Dummy Fitting")

    def test_should_return_eve_types(self):
        # given
        fit = create_fitting()
        # when
        types = fit.eve_types()
        # then
        self.assertSetEqual(
            {obj.id for obj in types}, {185, 31716, 3244, 593, 4405, 2873, 2205}
        )



# Generated by Django 4.0.10 on 2023-12-19 16:03

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("memberaudit", "0017_characterasset_bigauto_id"),
    ]

    operations = [
        migrations.CreateModel(
            name="CharacterCloneInfo",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("last_clone_jump_date", models.DateTimeField(default=None, null=True)),
                (
                    "last_station_change_date",
                    models.DateTimeField(default=None, null=True),
                ),
                (
                    "character",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="clone_info",
                        to="memberaudit.character",
                    ),
                ),
                (
                    "home_location",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="memberaudit.location",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
    ]



import os
import sys
from pathlib import Path
from string import Template


def main():
    myauth_dir = Path(__file__).parent.parent.parent.parent / "myauth"
    sys.path.insert(0, str(myauth_dir))

    import django

    # init and setup django project
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
    print("AA starting up...")
    django.setup()

    """MAIN"""
    from memberaudit.models import Character

    template = Template(
        """
        @shared_task(**_task_params)
        @when_esi_is_available
        def update_character_${value}(
            character_pk: int,
            force_update: bool,
            root_task_id: Optional[str] = None,
            parent_task_id: Optional[str] = None,
        ) -> None:
            \"\"\"Update ${value} for a character from ESI.\"\"\"
            _update_character_section(
                character_pk=character_pk,
                section=Character.UpdateSection.${name},
                force_update=force_update,
                root_task_id=root_task_id,
                parent_task_id=parent_task_id,
            )
    """
    )

    sections = set(Character.UpdateSection)
    excludes = {
        Character.UpdateSection.ASSETS,
        Character.UpdateSection.MAILS,
        Character.UpdateSection.CONTACTS,
        Character.UpdateSection.CONTRACTS,
    }
    for section in sorted(sections - excludes):
        print(template.substitute(name=section.name, value=section.value))


if __name__ == "__main__":
    main()



"""Template tags for Member Audit."""

from django import template
from django.urls import reverse

from memberaudit.models import Character

register = template.Library()


@register.simple_tag
def navactive_2(request, url_name: str, *args):
    """Return the active class name for navs."""
    url = reverse(url_name, args=args)
    if request.path == url:
        return "active"
    return ""


@register.inclusion_tag(
    "memberaudit/partials/character_viewer/tab_status_indicator.html",
    takes_context=True,
)
def tab_status_indicator(context, *sections) -> dict:
    """Render status indicator for a character tab.

    Show as error when at least one section has an error.

    Expects these keys in the context: "sections_update_status", "total_update_status"
    """
    sections_update_status = context["sections_update_status"]
    result = {"tab_update_status": Character.TotalUpdateStatus.OK}

    if context["total_update_status"] is Character.TotalUpdateStatus.DISABLED:
        return result

    is_success = True
    is_complete = True
    for section in sections:
        section_obj = Character.UpdateSection(section)  # make sure section is valid
        try:
            update_section = sections_update_status[str(section_obj)]
        except KeyError:
            is_complete = False
        else:
            if update_section.is_success is not None:
                is_success &= update_section.is_success

    if not is_success:
        tab_update_status = Character.TotalUpdateStatus.ERROR
    elif not is_complete:
        tab_update_status = Character.TotalUpdateStatus.INCOMPLETE
    else:
        tab_update_status = Character.TotalUpdateStatus.OK
    result["tab_update_status"] = tab_update_status
    return result


@register.inclusion_tag("memberaudit/partials/natural_number.html", takes_context=False)
def natural_number(value) -> dict:
    """Render a number humanized and with the exact number as tooltip."""
    return {"value": value}



from unittest.mock import patch

from django.test import TestCase

from app_utils.esi_testing import EsiClientStub, EsiEndpoint

from memberaudit.core.player_count import _fetch_player_count, clear_cache, get

MODULE_PATH = "memberaudit.core.player_count"


@patch(MODULE_PATH + ".esi")
class TestPlayerCount(TestCase):
    def test_should_return_player_count_when_available(self, mock_esi):
        # given
        endpoints = [
            EsiEndpoint(
                "Status",
                "get_status",
                data={
                    "players": 12345,
                    "server_version": "1132976",
                    "start_time": "2017-01-02T12:34:56Z",
                },
            )
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        clear_cache()

        # when
        result = get()

        # then
        self.assertEqual(result, 12345)

    def test_should_return_player_count_from_cache(self, mock_esi):
        # given
        endpoints = [
            EsiEndpoint(
                "Status",
                "get_status",
                data={
                    "players": 12345,
                    "server_version": "1132976",
                    "start_time": "2017-01-02T12:34:56Z",
                },
            )
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        clear_cache()
        get()

        # when
        with patch(
            MODULE_PATH + "._fetch_player_count", wraps=_fetch_player_count
        ) as spy:
            result = get()

            # then
            self.assertEqual(result, 12345)
            self.assertFalse(spy.called)

    def test_should_return_none_when_esi_offline(self, mock_esi):
        # given
        endpoints = [EsiEndpoint("Status", "get_status", http_error_code=500)]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        clear_cache()

        # when
        result = get()

        # then
        self.assertIsNone(result)



from django.test import RequestFactory, TestCase
from django.urls import reverse

from memberaudit.models import Character, CharacterUpdateStatus
from memberaudit.templatetags.memberaudit import navactive_2, tab_status_indicator


class TestNavactive2(TestCase):
    def setUp(self) -> None:
        self.factory = RequestFactory()

    def test_simple_return_active_when_matches(self):
        request = self.factory.get(reverse("memberaudit:add_character"))
        result = navactive_2(request, "memberaudit:add_character")
        self.assertEqual(result, "active")

    def test_simple_return_empty_when_no_match(self):
        request = self.factory.get(reverse("memberaudit:add_character"))
        result = navactive_2(request, "memberaudit:reports")
        self.assertEqual(result, "")

    def test_complex_return_active_when_matches(self):
        request = self.factory.get(reverse("memberaudit:character_viewer", args=[2]))
        result = navactive_2(request, "memberaudit:character_viewer", 2)
        self.assertEqual(result, "active")


class TestTabStatusIndicator(TestCase):
    def test_should_not_report_error_when_section_ok(self):
        # given
        status = CharacterUpdateStatus(
            section=Character.UpdateSection.LOCATION, is_success=True
        )
        sections_update_status = {"location": status}
        context = {
            "sections_update_status": sections_update_status,
            "total_update_status": Character.TotalUpdateStatus.OK,
        }
        # when
        result = tab_status_indicator(context, Character.UpdateSection.LOCATION)
        # then
        self.assertEqual(result["tab_update_status"], Character.TotalUpdateStatus.OK)

    def test_should_report_incomplete_when_section_not_found(self):
        # given
        status = CharacterUpdateStatus(
            section=Character.UpdateSection.MAILS, is_success=True
        )
        sections_update_status = {"mails": status}
        context = {
            "sections_update_status": sections_update_status,
            "total_update_status": Character.TotalUpdateStatus.OK,
        }
        # when
        result = tab_status_indicator(context, Character.UpdateSection.LOCATION)
        # then
        self.assertEqual(
            result["tab_update_status"], Character.TotalUpdateStatus.INCOMPLETE
        )

    def test_should_report_error_when_section_not_ok(self):
        # given
        status = CharacterUpdateStatus(
            section=Character.UpdateSection.LOCATION, is_success=False
        )
        sections_update_status = {"location": status}
        context = {
            "sections_update_status": sections_update_status,
            "total_update_status": Character.TotalUpdateStatus.ERROR,
        }
        # when
        result = tab_status_indicator(context, Character.UpdateSection.LOCATION)
        # then
        self.assertEqual(result["tab_update_status"], Character.TotalUpdateStatus.ERROR)

    def test_should_report_error_when_one_of_several_sections_not_ok(self):
        # given
        status_location = CharacterUpdateStatus(
            section=Character.UpdateSection.LOCATION, is_success=True
        )
        status_mails = CharacterUpdateStatus(
            section=Character.UpdateSection.LOCATION, is_success=False
        )
        sections_update_status = {"location": status_location, "mails": status_mails}
        context = {
            "sections_update_status": sections_update_status,
            "total_update_status": Character.TotalUpdateStatus.ERROR,
        }
        # when
        result = tab_status_indicator(
            context, Character.UpdateSection.LOCATION, Character.UpdateSection.MAILS
        )
        # then
        self.assertEqual(result["tab_update_status"], Character.TotalUpdateStatus.ERROR)

    def test_should_not_report_error_when_all_sections_ok(self):
        # given
        status_location = CharacterUpdateStatus(
            section=Character.UpdateSection.LOCATION, is_success=True
        )
        status_mails = CharacterUpdateStatus(
            section=Character.UpdateSection.LOCATION, is_success=True
        )
        sections_update_status = {"location": status_location, "mails": status_mails}
        context = {
            "sections_update_status": sections_update_status,
            "total_update_status": Character.TotalUpdateStatus.OK,
        }
        # when
        result = tab_status_indicator(
            context, Character.UpdateSection.LOCATION, Character.UpdateSection.MAILS
        )
        # then
        self.assertEqual(result["tab_update_status"], Character.TotalUpdateStatus.OK)

    def test_should_break_when_section_is_invalid(self):
        # given
        status_location = CharacterUpdateStatus(
            section=Character.UpdateSection.LOCATION, is_success=True
        )
        sections_update_status = {"location": status_location}
        context = {
            "sections_update_status": sections_update_status,
            "total_update_status": Character.TotalUpdateStatus.OK,
        }
        # when/then
        with self.assertRaises(ValueError):
            tab_status_indicator(context, "invalid-section-name")

    def test_should_not_report_error_when_status_is_disabled(self):
        # given
        status = CharacterUpdateStatus(
            section=Character.UpdateSection.LOCATION, is_success=False
        )
        sections_update_status = {"location": status}
        context = {
            "sections_update_status": sections_update_status,
            "total_update_status": Character.TotalUpdateStatus.DISABLED,
        }
        # when
        result = tab_status_indicator(context, Character.UpdateSection.LOCATION)
        # then
        self.assertEqual(result["tab_update_status"], Character.TotalUpdateStatus.OK)

    def test_should_report_incomplete_when_one_of_several_sections_is_missing(self):
        # given
        status_location = CharacterUpdateStatus(
            section=Character.UpdateSection.LOCATION, is_success=True
        )
        sections_update_status = {"location": status_location}
        context = {
            "sections_update_status": sections_update_status,
            "total_update_status": Character.TotalUpdateStatus.INCOMPLETE,
        }
        # when
        result = tab_status_indicator(
            context, Character.UpdateSection.LOCATION, Character.UpdateSection.MAILS
        )
        # then
        self.assertEqual(
            result["tab_update_status"], Character.TotalUpdateStatus.INCOMPLETE
        )

    def test_should_ignore_sections_in_progress(self):
        # given
        status_location = CharacterUpdateStatus(
            section=Character.UpdateSection.LOCATION, is_success=True
        )
        status_mails = CharacterUpdateStatus(
            section=Character.UpdateSection.LOCATION, is_success=None
        )
        sections_update_status = {"location": status_location, "mails": status_mails}
        context = {
            "sections_update_status": sections_update_status,
            "total_update_status": Character.TotalUpdateStatus.ERROR,
        }
        # when
        result = tab_status_indicator(
            context, Character.UpdateSection.LOCATION, Character.UpdateSection.MAILS
        )
        # then
        self.assertEqual(result["tab_update_status"], Character.TotalUpdateStatus.OK)



"""Helpers for Member Audit."""

import datetime as dt
import itertools
from typing import Any, Iterable, NamedTuple, Optional, Set

from celery import Task

from django.apps import apps
from django.forms.models import model_to_dict
from django.utils.timezone import now
from eveuniverse.models import EveEntity, EveType

from app_utils.datetime import datetime_round_hour

from memberaudit.app_settings import MEMBERAUDIT_DATA_RETENTION_LIMIT
from memberaudit.constants import EveDogmaAttributeId


class EveEntityIdsMixin:
    """Add EveEntity related features."""

    def eve_entity_ids(self) -> Set[int]:
        """Return IDs of all existing EveEntity objects for this object."""
        ids = set()
        for field in self._meta.get_fields():
            if field.is_relation and issubclass(field.related_model, EveEntity):
                value = getattr(self, field.attname)
                if value:
                    ids.add(value)
        return ids


def eve_entity_ids_from_objs(objs: Iterable[Any]) -> Set[int]:
    """Return all EveEntity IDs from objs. Will return an empty set when objs is empty.

    Expects objs to have the `EveEntityIdsMixin`.
    """
    if not objs:
        return set()

    entity_ids_list = [obj.eve_entity_ids() for obj in objs]
    return set(itertools.chain.from_iterable(entity_ids_list))


def data_retention_cutoff() -> Optional[dt.datetime]:
    """Return cutoff datetime for data retention of None if unlimited."""
    if MEMBERAUDIT_DATA_RETENTION_LIMIT is None:
        return None
    return datetime_round_hour(
        now() - dt.timedelta(days=MEMBERAUDIT_DATA_RETENTION_LIMIT)
    )


def implant_slot_num(implant_type: EveType) -> int:  # TODO: Refactor into model
    """Return slot number for an implant. Or 0 if not found."""
    dogma_attributes = {
        obj.eve_dogma_attribute_id: obj.value
        for obj in implant_type.dogma_attributes.all()
    }
    try:
        slot_num = int(dogma_attributes[EveDogmaAttributeId.IMPLANT_SLOT])
    except KeyError:
        slot_num = 0
    return slot_num


def determine_task_priority(task_obj: Task) -> Optional[int]:
    """Return priority of give task or None if not defined."""
    properties = task_obj.request.get("properties") or {}
    return properties.get("priority")


def arabic_number_to_roman(value) -> str:
    """Map to convert arabic to roman numbers (1 to 5 only)"""
    my_map = {0: "-", 1: "I", 2: "II", 3: "III", 4: "IV", 5: "V"}
    try:
        return my_map[value]
    except KeyError:
        return "-"


class UpdateSectionResult(NamedTuple):
    """A result of an attempted section update."""

    is_changed: Optional[bool]
    is_updated: bool
    data: Any = None


def character_section_models():
    """Return all character section models."""
    my_app = apps.get_app_config("memberaudit")
    my_character_models = [
        model_class
        for model_class in my_app.get_models()
        if model_class.__name__.startswith("Character")
    ]

    return my_character_models


def model_to_dict_safely(obj) -> dict:
    """Convert a model ot dict in a safe manner."""
    fields = [field.name for field in obj._meta.fields]
    result = model_to_dict(obj, fields=fields)
    return result



# Generated by Django 4.2.16 on 2024-10-07 19:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("memberaudit", "0019_alter_charactercontractitem_record_id"),
    ]

    operations = [
        migrations.AlterField(
            model_name="characterrole",
            name="role",
            field=models.CharField(
                choices=[
                    ("AT1", "account take 1"),
                    ("AT2", "account take 2"),
                    ("AT3", "account take 3"),
                    ("AT4", "account take 4"),
                    ("AT5", "account take 5"),
                    ("AT6", "account take 6"),
                    ("AT7", "account take 7"),
                    ("ACT", "accountant"),
                    ("AUD", "auditor"),
                    ("BMG", "brand manager"),
                    ("COM", "communications officer"),
                    ("CEQ", "config equipment"),
                    ("CSE", "config starbase equipment"),
                    ("CT1", "container take 1"),
                    ("CT2", "container take 2"),
                    ("CT3", "container take 3"),
                    ("CT4", "container take 4"),
                    ("CT5", "container take 5"),
                    ("CT6", "container take 6"),
                    ("CT7", "container take 7"),
                    ("CMG", "contract manager"),
                    ("DCT", "deliveries container take"),
                    ("DQY", "deliveries query"),
                    ("DTK", "deliveries take"),
                    ("DPL", "diplomat"),
                    ("DRT", "director"),
                    ("FCM", "factory manager"),
                    ("FTM", "fitting manager"),
                    ("HQ1", "hangar query 1"),
                    ("HQ2", "hangar query 2"),
                    ("HQ3", "hangar query 3"),
                    ("HQ4", "hangar query 4"),
                    ("HQ5", "hangar query 5"),
                    ("HQ6", "hangar query 6"),
                    ("HQ7", "hangar query 7"),
                    ("HT1", "hangar take 1"),
                    ("HT2", "hangar take 2"),
                    ("HT3", "hangar take 3"),
                    ("HT4", "hangar take 4"),
                    ("HT5", "hangar take 5"),
                    ("HT6", "hangar take 6"),
                    ("HT7", "hangar take 7"),
                    ("JAC", "junior accountant"),
                    ("PSM", "personnel manager"),
                    ("PMG", "project manager"),
                    ("RFF", "rent factory facility"),
                    ("RFC", "rent office"),
                    ("RRF", "rent research facility"),
                    ("SCO", "security officer"),
                    ("SPM", "skill plan manager"),
                    ("SDO", "starbase defense operator"),
                    ("SFT", "starbase fuel technician"),
                    ("STM", "station manager"),
                    ("TRD", "trader"),
                ],
                help_text="Role of a character",
                max_length=3,
            ),
        ),
    ]



"""Shared utils for tests."""

import json
import logging
from typing import Any, Tuple

from django.contrib.auth.models import Permission, User
from django.db.models import QuerySet
from django.http import JsonResponse
from django.test import TestCase
from esi.models import Token

from allianceauth.authentication.backends import StateBackend
from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.allianceauth import get_redis_client
from app_utils.testing import NoSocketsTestCase, add_character_to_user, response_text

from memberaudit.models import Character

from .testdata.factories import create_character

logger = logging.getLogger(__name__)


def create_user_from_evecharacter_with_access(
    character_id: int, disconnect_signals: bool = True
) -> Tuple[User, CharacterOwnership]:
    """Create user with access from an existing eve character and use it as main."""
    auth_character = EveCharacter.objects.get(character_id=character_id)
    username = StateBackend.iterate_username(auth_character.character_name)
    user = AuthUtils.create_user(username, disconnect_signals=disconnect_signals)
    user = AuthUtils.add_permission_to_user_by_name(
        "memberaudit.basic_access", user, disconnect_signals=disconnect_signals
    )
    character_ownership = add_character_to_user(
        user,
        auth_character,
        is_main=True,
        scopes=Character.get_esi_scopes(),
        disconnect_signals=disconnect_signals,
    )
    return user, character_ownership


def create_memberaudit_character(
    character_id: int, disconnect_signals: bool = True, **kwargs
) -> Character:
    """Create a memberaudit character from an existing auth character
    incl. user and making it the main.
    """
    _, character_ownership = create_user_from_evecharacter_with_access(
        character_id, disconnect_signals=disconnect_signals
    )
    return create_character(eve_character=character_ownership.character, **kwargs)


def add_auth_character_to_user(
    user: User, character_id: int, scopes=None, disconnect_signals=True
) -> CharacterOwnership:
    auth_character = EveCharacter.objects.get(character_id=character_id)
    if not scopes:
        scopes = Character.get_esi_scopes()

    return add_character_to_user(
        user,
        auth_character,
        is_main=False,
        scopes=scopes,
        disconnect_signals=disconnect_signals,
    )


def add_memberaudit_character_to_user(
    user: User, character_id: int, disconnect_signals: bool = True, **kwargs
) -> Character:
    character_ownership = add_auth_character_to_user(
        user, character_id, disconnect_signals=disconnect_signals
    )
    return create_character(eve_character=character_ownership.character, **kwargs)


def scope_names_set(token: Token) -> set:
    return set(token.scopes.values_list("name", flat=True))


def json_response_to_python_2(response: JsonResponse, data_key="data") -> Any:
    """Convert JSON response into Python object."""
    data = json.loads(response_text(response))
    return data[data_key]


def json_response_to_dict_2(response: JsonResponse, key="id", data_key="data") -> dict:
    """Convert JSON response into dict by given key."""
    return {o[key]: o for o in json_response_to_python_2(response, data_key)}


class TestCaseWithFixtures(TestCase):
    fixtures = ["disable_analytics.json"]


class NoSocketsTestCaseFixtures(NoSocketsTestCase):
    fixtures = ["disable_analytics.json"]


def permissions_for_model(model_class) -> QuerySet:
    """Return all permissions defined for a model."""
    app_label = model_class._meta.app_label
    model_name = model_class._meta.model_name
    return Permission.objects.filter(
        content_type__app_label=app_label, content_type__model=model_name
    )


def reset_celery_once_locks():
    """Reset celery once locks for given tasks."""
    r = get_redis_client()
    app_label = "memberaudit"
    if keys := r.keys(f":?:qo_{app_label}.*"):
        deleted_count = r.delete(*keys)
        logger.info("Removed %d stuck celery once keys", deleted_count)
    else:
        deleted_count = 0



import logging

from django.core.management import call_command
from django.core.management.base import BaseCommand

from app_utils.logging import LoggerAddTag

from memberaudit import __title__
from memberaudit.constants import EveCategoryId

logger = LoggerAddTag(logging.getLogger(__name__), __title__)


class Command(BaseCommand):
    help = "Preloads data required for this app from ESI"

    def handle(self, *args, **options):
        call_command(
            "eveuniverse_load_types",
            __title__,
            "--category_id_with_dogma",
            str(EveCategoryId.BLUEPRINT.value),
            "--category_id_with_dogma",
            str(EveCategoryId.SHIP.value),
            "--category_id_with_dogma",
            str(EveCategoryId.MODULE.value),
            "--category_id_with_dogma",
            str(EveCategoryId.CHARGE.value),
            "--category_id_with_dogma",
            str(EveCategoryId.SKILL.value),
            "--category_id_with_dogma",
            str(EveCategoryId.DRONE.value),
            "--category_id_with_dogma",
            str(EveCategoryId.IMPLANT.value),
            "--category_id_with_dogma",
            str(EveCategoryId.FIGHTER.value),
            "--category_id_with_dogma",
            str(EveCategoryId.SUBSYSTEM.value),
            "--category_id",
            str(EveCategoryId.STRUCTURE.value),
        )



"""An Alliance Auth app that provides full access to Eve characters
and related reports for auditing, vetting and monitoring.
"""

# pylint: disable=invalid-name
default_app_config = "memberaudit.apps.MemberAuditConfig"

__version__ = "3.12.1"
__title__ = "Member Audit"



"""Provide cached access to the current player count of the Eve server."""

from typing import Optional

from django.core.cache import cache

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from memberaudit import __title__
from memberaudit.providers import esi

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

_CACHE_KEY = "memberaudit-player-count"
_TIMEOUT = 1 * 60


def get() -> Optional[int]:
    """Return cached player count from ESI or None if offline."""
    return cache.get_or_set(
        key=_CACHE_KEY, default=_fetch_player_count, timeout=_TIMEOUT
    )


def clear_cache():
    """Clear cache."""
    cache.delete(_CACHE_KEY)


def _fetch_player_count() -> Optional[int]:
    try:
        result: dict = esi.client.Status.get_status().results()
    except OSError:
        return None

    return result.get("players")
